/*
    Main class for running the CBR-cycle.
    Loads the case bases and runs each step of the CBR-cycle.
 */

import com.mongodb.MongoClient;
import com.mongodb.client.MongoCollection;
import com.mongodb.client.MongoCursor;
import com.mongodb.client.MongoDatabase;
import org.bson.Document;
import org.json.JSONException;

import java.io.*;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;

public class CBRMain {

    //Creates the ArrayLists to hold the equipment, squads and armies.
    public static ArrayList<Equipment> EquipmentList;
    public static ArrayList<Squad> SquadList;
    public static ArrayList<Army> ArmyList;

    //Global Weights
    public static double movementWeight;
    public static double psychicWeight;
    public static double shootingWeight;
    public static double assaultWeight;
    public static int move = 0;
    public static int psy = 0;
    public static int shoot = 0;
    public static int melee = 0;

    //Variables used in utility maintenance
    public static long retrieveTime = 0;
    public static boolean retrievalTimeTrue = false;

    //Creates lists to hold parameters and options, necessary for retention.
    public static List<List<List<Integer>>> options = new ArrayList<>();
    public static List<List<List<String>>> parameters = new ArrayList<>();
    public static List<List<Integer>> squadOptions = new ArrayList<>();
    public static List<List<String>> squadParameters = new ArrayList<>();

    //Creates an explanation list to hold explanations
    public static List<String> retrievalExplanations = new ArrayList<>();
    public static List<String> reuseExplanations = new ArrayList<>();
    public static List<String> reviseExplanations = new ArrayList<>();
    public static List<String> retainExplanations = new ArrayList<>();
    public static List<String> maintenanceExplanations = new ArrayList<>();

    //Other important variables for explanations
    public static double winChanceBeforeReuse = 0;
    public static double winChanceAfterReuse = 0;
    public static double winChanceAfterRevise = 0;

    //The Main Class starts up the database and the processes.
    public static void main(String[] args) throws JSONException, ParseException, IOException {

        //Load in the weights from notepad file with weights
        BufferedReader in = new BufferedReader(new FileReader("weights.txt"));
        movementWeight = Double.parseDouble(in.readLine());
        psychicWeight = Double.parseDouble(in.readLine());
        shootingWeight = Double.parseDouble(in.readLine());
        assaultWeight = Double.parseDouble(in.readLine());
        in.close();

        //Below is the connection to a Mongo Database, where the Case Base is stored.
        MongoClient mongoClient = new MongoClient();
        MongoDatabase db = mongoClient.getDatabase("local");

        //Load equipment from casebase.
        MongoCollection<Document> collection = db.getCollection("equipment");
        MongoCursor<Document> cursor = collection.find().iterator();
        EquipmentList = CaseBaseLoading.loadEquipment(collection, cursor);

        //Load squads from casebase.
        collection = db.getCollection("unit");
        cursor = collection.find().iterator();
        SquadList = CaseBaseLoading.loadSquads(collection, cursor);

        //Loads armies from casebase
        collection = db.getCollection("army");
        cursor = collection.find().iterator();
        ArmyList = CaseBaseLoading.loadArmies(collection, cursor);

        /*
            The CBR-cycle starts here
         */

        //Assign problem army.
        Army problemArmy = ArmyList.get(0);

        //Retrieve the best army.
        Army best = Retrieve.RetrieveBest(ArmyList, problemArmy);

        //Copy the retrieved army, we do not want it to be changed.
        List<Squad> bestList = new ArrayList<>();
        for (int k = 0; k < best.getSquad().size(); k++) {
            bestList.add(new Squad(best.getSquad().get(k)));
        }

        Army b = new Army(best.getName(), best.getCost(), best.getRating(), bestList);

        //Copy the list of options and parameters, so that they are specific to the solution army
        for (int i = 0; i < CBRMain.ArmyList.size(); i++) {
            List<List<Integer>> optTmpList = new ArrayList<>();
            List<List<String>> parTmpList = new ArrayList<>();
            if (best.getName().equals(CBRMain.ArmyList.get(i).getName())) {
                for (int j = 0; j < options.get(i).size(); j++) {
                    optTmpList.add(new ArrayList<>(options.get(i).get(j)));
                    parTmpList.add(new ArrayList<>(parameters.get(i).get(j)));
                }
                options.add(optTmpList);
                parameters.add(parTmpList);
            }
        }

        //First Explanation

        System.out.println(retrievalExplanations.get(0));
        retrievalExplanations.remove(0);

        System.out.println();
        System.out.println("The best retrieved army is named: " + best.getName() + " in the system.");
        System.out.println("It has a rating of: " + best.getRating());
        System.out.println();
        System.out.println("The name of each squad, its options, parameters and the ratings are presented below.");
        System.out.println();
        for (int i = 0; i < best.getSquad().size(); i++){
            System.out.println("Squad named: " + best.getSquad().get(i).getName() + " in the system has a rating of " + best.getSquad().get(i).getRating());
            System.out.println("It has the following options applied: " + options.get(options.size()-1).get(i) + " with the following parameters " + parameters.get(parameters.size()-1).get(i));
            System.out.println();
        }

        Scanner input = new Scanner(System.in);
        String s = "";
        while (!s.equals("Y") && !s.equals("N") && !s.equals("y") && !s.equals("n")) {
            System.out.println("Would you like a more in depth explanation? Type Y for Yes, and N for No.");
            s = input.nextLine();
        }

        if (s.equals("y") || s.equals("Y")) {
            for (int i = 0; i < retrievalExplanations.size(); i++) {
                System.out.println(retrievalExplanations.get(i));
            }
        }

        long reuseTime = System.currentTimeMillis();
        //Reuse (Adapt) the best army.
        b = Reuse.mainReuse(b, problemArmy);
        reuseTime = System.currentTimeMillis() - reuseTime;

        //Second Explanation
        System.out.println(reuseExplanations.get(0));
        if (winChanceBeforeReuse != winChanceAfterReuse) {
            System.out.println();
            System.out.println("After adapting the army, the following army composition was created.");
            System.out.println();
            System.out.println("The name of each squad, its options, parameters and the ratings are presented below.");
            System.out.println();
            for (int i = 0; i < b.getSquad().size(); i++) {
                System.out.println("Squad named: " + b.getSquad().get(i).getName() + " in the system has a rating of " + b.getSquad().get(i).getRating());
                System.out.println("It has the following options applied: " + options.get(options.size() - 1).get(i) + " with the following parameters " + parameters.get(parameters.size() - 1).get(i));
                System.out.println();
            }
        }
        reuseExplanations.remove(0);

        s = "";
        if (!reuseExplanations.isEmpty()) {
            while (!s.equals("Y") && !s.equals("N") && !s.equals("y") && !s.equals("n")) {
                System.out.println("Would you like a more in depth explanation? Type Y for Yes, and N for No.");
                s = input.nextLine();
            }


            if (s.equals("y") || s.equals("Y")) {
                for (int i = 0; i < reuseExplanations.size(); i++) {
                    System.out.println(reuseExplanations.get(i));
                }
            }
        }

        //Revise the best army, after it had been used in practice.
        b = Revise.mainRevise(b, problemArmy);

        //Third Explanation

        System.out.println(reviseExplanations.get(0));
        System.out.println();
        reviseExplanations.remove(0);
        System.out.println(reviseExplanations.get(0));
        System.out.println();
        reviseExplanations.remove(0);
        System.out.println(reviseExplanations.get(0));
        if (winChanceAfterReuse != winChanceAfterRevise) {
            System.out.println();
            System.out.println("After revising the army, the following army composition was created.");
            System.out.println();
            System.out.println("The name of each squad, its options, parameters and the ratings are presented below.");
            System.out.println();
            for (int i = 0; i < b.getSquad().size(); i++) {
                System.out.println("Squad named: " + b.getSquad().get(i).getName() + " in the system has a rating of " + b.getSquad().get(i).getRating());
                System.out.println("It has the following options applied: " + options.get(options.size() - 1).get(i) + " with the following parameters " + parameters.get(parameters.size() - 1).get(i));
                System.out.println();
            }
        }
        reviseExplanations.remove(0);

        s = "";
        while (!s.equals("Y") && !s.equals("N") && !s.equals("y") && !s.equals("n")) {
            System.out.println("Would you like a more in depth explanation? Type Y for Yes, and N for No.");
            s = input.nextLine();
        }

        if (s.equals("y") || s.equals("Y")) {
            for (int i = 0; i < reviseExplanations.size(); i++) {
                System.out.println(reviseExplanations.get(i));
            }
        }

        //Retain the armies, squads and the like into the system.
        Retain.mainRetain(b, problemArmy, db);

        //Fourth Explanation
        System.out.println(CBRMain.retainExplanations.get(0));
        retainExplanations.remove(0);
        System.out.println(CBRMain.retainExplanations.get(0));
        retainExplanations.remove(0);
        s = "";
        while (!s.equals("Y") && !s.equals("N") && !s.equals("y") && !s.equals("n")) {
            System.out.println("Would you like a more in depth list of saved armies and their ratings? Type Y for Yes, and N for No.");
            s = input.nextLine();
        }

        if (s.equals("y") || s.equals("Y")) {
            for (int i = 0; i < retainExplanations.size(); i++) {
                System.out.println(retainExplanations.get(i));
            }
        }

        double utilityRatio = (retrieveTime*1.0)/(reuseTime*1.0);

        if (utilityRatio >= 0.9){
            /*
                If the utility ratio is higher than 0.9, we perform the utility maintenance.
                First we take a look at all the armies, and we delete the lowest ranking army.
                N represents the lowest amount of armies we should have of one faction.
                M represents the lowest amount of none base squads we should have of one faction.
             */

            int n = 4;
            int m = 8;

            maintenanceExplanations.add("The performance of the system has reached a critical point. Maintenance of the case base will be started.");
            maintenanceExplanations.add("The system will keep a minimum of " + n + " armies of each faction, and " + m + " non-basic squads.");
            maintenanceExplanations.add("The rest will be deleted, lowest rank first.");

            System.out.println(maintenanceExplanations.get(0));
            maintenanceExplanations.remove(0);
            System.out.println(maintenanceExplanations.get(0));
            maintenanceExplanations.remove(0);
            System.out.println(maintenanceExplanations.get(0));
            maintenanceExplanations.remove(0);

            //We count army factions and store them in a list.
            List<Integer> armyCount = UtilityMaintenanceFunctions.countArmies();

            //Now we delete candidates
            for (int i = 0; i < armyCount.size(); i++){
                while (armyCount.get(i) > n) {
                    UtilityMaintenanceFunctions.removeLowest(db, i);
                    armyCount.set(i, armyCount.get(i)-1);
                }
            }

            //Same for squads, first we count factions.
            List<Integer> squadCount = UtilityMaintenanceFunctions.countSquads();

            //Then we delete candidates
            for (int i = 0; i < squadCount.size(); i++){
                while (squadCount.get(i) > m) {
                    UtilityMaintenanceFunctions.removeLowestSquad(db, i);
                    squadCount.set(i, squadCount.get(i)-1);
                }
            }

            s = "";
            while (!s.equals("Y") && !s.equals("N") && !s.equals("y") && !s.equals("n")) {
                System.out.println("Would you like a more in depth list of removed armies? Type Y for Yes, and N for No.");
                s = input.nextLine();
            }

            if (s.equals("y") || s.equals("Y")) {
                for (int i = 0; i < maintenanceExplanations.size(); i++) {
                    System.out.println(maintenanceExplanations.get(i));
                }
            }

        }

        for (int i = 0; i < 5; i++) {
            MetagameMaintenance.runMaintenance(db);
        }

        while(move > 10){
            move-=10;
            movementWeight+=0.01;
        }

        while(move < -10){
            move+=10;
            movementWeight-=0.01;
        }

        while(psy > 10){
            psy-=10;
            psychicWeight+=0.01;
        }

        while(psy < -10){
            psy+=10;
            psychicWeight-=0.01;
        }

        while(shoot > 10){
            shoot-=10;
            shootingWeight+=0.01;
        }

        while(shoot < -10){
            shoot+=10;
            shootingWeight-=0.01;
        }

        while(melee > 10){
            melee-=10;
            assaultWeight+=0.01;
        }

        while(melee < -10){
            melee+=10;
            assaultWeight-=0.01;
        }

        //Write the weights back to the .txt file, in case they changed.
        BufferedWriter out = new BufferedWriter(new FileWriter("weights.txt"));
        out.write(Double.toString(movementWeight));
        out.newLine();
        out.write(Double.toString(psychicWeight));
        out.newLine();
        out.write(Double.toString(shootingWeight));
        out.newLine();
        out.write(Double.toString(assaultWeight));
        out.close();

        //Close the client when we are done with using the system.
        mongoClient.close();
    }
}