/*
    This class loads the entire case base. It is a helper class to the CBRMain, so that the CBRMain would not be cluttered.
 */

import com.mongodb.client.MongoCollection;
import com.mongodb.client.MongoCursor;
import org.bson.Document;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.util.ArrayList;
import java.util.List;

public class CaseBaseLoading {

    //This method loads the equipment into the database.
    public static ArrayList<Equipment> loadEquipment(MongoCollection<Document> collection, MongoCursor<Document> cursor) throws JSONException {

        ArrayList<Equipment> EquipmentList = new ArrayList<>();

        String name;
        String strength;
        String json;
        double range;
        double ap;

        Document doc;

        for (int i = 0; i < collection.count(); i++){

            //Set all of the variables to their respective values.
            doc = cursor.next();
            name = doc.getString("Name");
            range = doc.getDouble("Range");
            strength = doc.getString("Strength");
            ap = doc.getDouble("AP");
            //Set the document to JSON. Now we extract the JSON Arrays.
            json = doc.toJson();

            //Takes the object and makes it a JSON Object.
            JSONObject obj = new JSONObject(json);

            //Creates a list and places the JSON array values inside this list.
            List<String> list = new ArrayList<>();
            JSONArray array = obj.getJSONArray("Type");
            for(int j = 0 ; j < array.length() ; j++){
                list.add(array.getJSONObject(j).getString("Name"));
            }

            Equipment equipment = new Equipment(name, range, strength, ap, list);
            EquipmentList.add(equipment);
        }

        return EquipmentList;
    }

    //This method loads the Squads into the case base.
    public static ArrayList<Squad> loadSquads(MongoCollection<Document> collection, MongoCursor<Document> cursor) throws JSONException {

        ArrayList<Squad> SquadList = new ArrayList<>();

        String name;
        String role;
        String unitname;
        String type;
        String faction;
        String trait;
        String json;

        double cost;
        boolean base;
        double rating;
        double BS;
        double WS;
        double S;
        double T;
        double W;
        double I;
        double A;
        double LD;
        double AS;
        boolean unique;
        double count;
        double front;
        double side;
        double rear;
        double hp;

        Document doc;

        for (int i = 0; i < collection.count(); i++){{

            //Set all of the variables to their respective values.
            doc = cursor.next();
            name = doc.getString("Name");
            cost = doc.getDouble("Cost");
            role = doc.getString("Role");
            base = doc.getBoolean("Base");
            rating = doc.getDouble("Rating");
            ArrayList<Unit> UnitList = new ArrayList<>();

            //Set the document to JSON. Now we extract the JSON Arrays.
            json = doc.toJson();

            //Takes the object and makes it a JSON Object.
            JSONObject obj = new JSONObject(json);

            //Creates units and places them in a list
            JSONArray array = obj.getJSONArray("Units");
            for(int j = 0 ; j < array.length() ; j++){
                //The unit can either be an infantry, walker or vehicle
                type = (array.getJSONObject(j).getString("Unit Type"));
                switch (type) {
                    case "Infantry":
                    case "Infantry-Character":
                    case "Bike":
                    case "Bike-Character":
                    case "Jump Unit":
                    case "Jetbike":
                    case "Jetbike-Character":
                    case "Artillery": {
                        List<UnitEq> equipment = new ArrayList<>();
                        List<String> rules = new ArrayList<>();
                        List<Integer> options = new ArrayList<>();
                        unitname = (array.getJSONObject(j).getString("Name"));
                        BS = (array.getJSONObject(j).getDouble("Ballistics Skill"));
                        WS = (array.getJSONObject(j).getDouble("Weapon Skill"));
                        S = (array.getJSONObject(j).getDouble("Strength"));
                        T = (array.getJSONObject(j).getDouble("Toughness"));
                        W = (array.getJSONObject(j).getDouble("Wounds"));
                        I = (array.getJSONObject(j).getDouble("Initiative"));
                        A = (array.getJSONObject(j).getDouble("Attacks"));
                        LD = (array.getJSONObject(j).getDouble("Leadership"));
                        AS = (array.getJSONObject(j).getDouble("Armour Save"));
                        unique = (array.getJSONObject(j).getBoolean("Unique"));
                        count = (array.getJSONObject(j).getDouble("Count"));
                        faction = (array.getJSONObject(j).getString("Faction"));
                        trait = (array.getJSONObject(j).getString("Warlord Trait"));
                        JSONArray eq = (array.getJSONObject(j).getJSONArray("Equipment"));
                        for (int k = 0; k < eq.length(); k++) {
                            equipment.add(new UnitEq(eq.getJSONObject(k).getString("Name"), eq.getJSONObject(k).getInt("Count")));
                        }
                        JSONArray r = (array.getJSONObject(j).getJSONArray("Special Rules"));
                        for (int k = 0; k < r.length(); k++) {
                            rules.add(r.getJSONObject(k).getString("Name"));
                        }
                        JSONArray o = (array.getJSONObject(j).getJSONArray("Options"));
                        for (int k = 0; k < o.length(); k++) {
                            options.add(o.getJSONObject(k).getInt("Name"));
                        }

                        Unit unit = new Unit(unitname, BS, WS, S, T, W, I, A, LD, AS, type, unique, count, faction, equipment, rules, options, trait);
                        UnitList.add(unit);
                        break;
                    }
                    case "Vehicle-Walker": {
                        List<UnitEq> equipment = new ArrayList<>();
                        List<String> rules = new ArrayList<>();
                        List<Integer> options = new ArrayList<>();
                        unitname = (array.getJSONObject(j).getString("Name"));
                        BS = (array.getJSONObject(j).getDouble("Ballistics Skill"));
                        WS = (array.getJSONObject(j).getDouble("Weapon Skill"));
                        S = (array.getJSONObject(j).getDouble("Strength"));
                        hp = (array.getJSONObject(j).getDouble("HP"));
                        I = (array.getJSONObject(j).getDouble("Initiative"));
                        A = (array.getJSONObject(j).getDouble("Attacks"));
                        front = (array.getJSONObject(j).getDouble("F"));
                        side = (array.getJSONObject(j).getDouble("S"));
                        rear = (array.getJSONObject(j).getDouble("R"));
                        unique = (array.getJSONObject(j).getBoolean("Unique"));
                        count = (array.getJSONObject(j).getDouble("Count"));
                        faction = (array.getJSONObject(j).getString("Faction"));
                        trait = (array.getJSONObject(j).getString("Warlord Trait"));
                        JSONArray eq = (array.getJSONObject(j).getJSONArray("Equipment"));
                        for (int k = 0; k < eq.length(); k++) {
                            equipment.add(new UnitEq(eq.getJSONObject(k).getString("Name"), eq.getJSONObject(k).getInt("Count")));
                        }
                        JSONArray r = (array.getJSONObject(j).getJSONArray("Special Rules"));
                        for (int k = 0; k < r.length(); k++) {
                            rules.add(r.getJSONObject(k).getString("Name"));
                        }
                        JSONArray o = (array.getJSONObject(j).getJSONArray("Options"));
                        for (int k = 0; k < o.length(); k++) {
                            options.add(o.getJSONObject(k).getInt("Name"));
                        }

                        Unit unit = new Unit(unitname, BS, S, WS, I, A, type, unique, count, faction, equipment, rules, options, trait, front, side, rear, hp);
                        UnitList.add(unit);
                        break;
                    }
                    default: {
                        List<UnitEq> equipment = new ArrayList<>();
                        List<String> rules = new ArrayList<>();
                        List<Integer> options = new ArrayList<>();
                        unitname = (array.getJSONObject(j).getString("Name"));
                        BS = (array.getJSONObject(j).getDouble("Ballistics Skill"));
                        hp = (array.getJSONObject(j).getDouble("HP"));
                        front = (array.getJSONObject(j).getDouble("F"));
                        side = (array.getJSONObject(j).getDouble("S"));
                        rear = (array.getJSONObject(j).getDouble("R"));
                        unique = (array.getJSONObject(j).getBoolean("Unique"));
                        count = (array.getJSONObject(j).getDouble("Count"));
                        faction = (array.getJSONObject(j).getString("Faction"));
                        trait = (array.getJSONObject(j).getString("Warlord Trait"));
                        JSONArray eq = (array.getJSONObject(j).getJSONArray("Equipment"));
                        for (int k = 0; k < eq.length(); k++) {
                            equipment.add(new UnitEq(eq.getJSONObject(k).getString("Name"), eq.getJSONObject(k).getInt("Count")));
                        }
                        JSONArray r = (array.getJSONObject(j).getJSONArray("Special Rules"));
                        for (int k = 0; k < r.length(); k++) {
                            rules.add(r.getJSONObject(k).getString("Name"));
                        }
                        JSONArray o = (array.getJSONObject(j).getJSONArray("Options"));
                        for (int k = 0; k < o.length(); k++) {
                            options.add(o.getJSONObject(k).getInt("Name"));
                        }
                        Unit unit = new Unit(unitname, BS, front, side, rear, hp, type, unique, count, faction, equipment, rules, options, trait);
                        UnitList.add(unit);
                        break;
                    }
                }
            }

            List<Integer> options = new ArrayList<>();
            List<String> parameters = new ArrayList<>();
            array = obj.getJSONArray("Options");
            for (int k = 0; k < array.length(); k++){
                options.add(array.getJSONObject(k).getInt("Name"));
            }
            array = obj.getJSONArray("Parameters");
            for (int k = 0; k < array.length(); k++){
                parameters.add(array.getJSONObject(k).getString("Name"));
            }
            CBRMain.squadOptions.add(options);
            CBRMain.squadParameters.add(parameters);

            Squad squad = new Squad(name, cost, role, base, rating, UnitList);
            SquadList.add(squad);
        }}

        return SquadList;
    }

    //This method loads the armies into the case base.
    public static ArrayList<Army> loadArmies(MongoCollection<Document> collection, MongoCursor<Document> cursor) throws JSONException {

        ArrayList<Army> ArmyList = new ArrayList<>();

        String name;
        String json;
        double cost;
        double rating;

        Document doc;

        for (int i = 0; i < collection.count(); i++){
            //Set all of the variables to their respective values.
            doc = cursor.next();
            name = doc.getString("Name");
            cost = doc.getDouble("Cost");
            rating = doc.getDouble("Rating");

            //Set the document to JSON. Now we extract the JSON Arrays.
            json = doc.toJson();

            //Takes the object and makes it a JSON Object.
            JSONObject obj = new JSONObject(json);

            //Creates a list and places the JSON array values inside this list.
            List<List<Integer>> optionsList = new ArrayList<>();
            List<List<String>> parameterList = new ArrayList<>();

            List<Squad> sq = new ArrayList<>();
            JSONArray array = obj.getJSONArray("Squads");
            for (int j = 0; j < array.length(); j++) {

                List<Integer> op = new ArrayList<>();
                List<String> par = new ArrayList<>();
                Squad s = null;

                //Adds squads to the squad list for a specific army.
                for (int k = 0; k < CBRMain.SquadList.size(); k++) {
                    if (CBRMain.SquadList.get(k).getName().equals(array.getJSONObject(j).getString("Name"))) {
                        s = new Squad(CBRMain.SquadList.get(k));
                        sq.add(s);
                    }
                }
                //Checks the options...
                JSONArray o = (array.getJSONObject(j).getJSONArray("Options"));
                for (int k = 0; k < o.length(); k++) {
                    op.add(o.getJSONObject(k).getInt("Option"));
                }
                //...and their parameters...
                JSONArray p = (array.getJSONObject(j).getJSONArray("Parameter"));
                for (int k = 0; k < p.length(); k++) {
                    par.add(p.getJSONObject(k).getString("Name"));
                }
                //And applies them to the squad (thus we don't have to store hundreds of lines for 1 army, we reuse the basic armies and apply options.
                for (int k = 0; k < op.size(); k++) {
                    GeneralKnowledge.Option(op.get(k), par.get(k), sq.get(sq.size() - 1));
                }

                //Add the armies squads to the Squad Case Base
                int countOccurences = 1;
                for (int k = 0; k < CBRMain.SquadList.size(); k++) {
                    if (CBRMain.SquadList.get(k).getName().contains(sq.get(sq.size() - 1).getName())) {
                        countOccurences++;
                    }
                }

                int countDuplicates = 0;

                //Make insert a new squad (so that the naming is preserved for the army).
                Squad insert = new Squad(sq.get(sq.size() - 1));

                //Set the name and the base of the insert to be different, so that it is a different case.
                insert.setName(sq.get(sq.size() - 1).getName() + " " + countOccurences);
                insert.setRating(1500.0);
                insert.setBase(false);

                for (int k = 0; k < CBRMain.SquadList.size(); k++) {
                    if (CBRMain.SquadList.get(k).getName().contains(sq.get(sq.size() - 1).getName())) {
                        //Check if there is a single duplicate in the system. If there is, we do not add a squad, instead we adjust it's rating.
                        if (checkDuplicate(CBRMain.SquadList.get(k), insert)) {
                            countDuplicates++;
                            if (s != null) {
                                s.setRating(CBRMain.SquadList.get(k).getRating());
                            }
                        }
                    }
                }

                if (countDuplicates == 0){
                    CBRMain.SquadList.add(insert);
                    CBRMain.squadOptions.add(op);
                    CBRMain.squadParameters.add(par);
                }

                optionsList.add(op);
                parameterList.add(par);
            }

            //The armies are added to the army list, ready to be reused.
            Army army = new Army(name, cost, rating, sq);
            ArmyList.add(army);

            CBRMain.options.add(optionsList);
            CBRMain.parameters.add(parameterList);
        }

        return ArmyList;
    }

    public static boolean checkDuplicate(Squad s, Squad t){
        if (s == null || t == null){
            return false;
        }
        //Check if the sizes of the squad are the same. Obviously, squads with transports and without are different.
        if (s.getUnit().size() != t.getUnit().size()){
            return false;
        }
        //Check if the names and counts of each unit are the same.
        for (int i = 0; i < s.getUnit().size(); i++){
            if (!s.getUnit().get(i).getName().equals(t.getUnit().get(i).getName()) || !(s.getUnit().get(i).getCount() == t.getUnit().get(i).getCount())){
                return false;
            }
        }
        // Check the equipment sizes of the squads.
        for (int i = 0; i < s.getUnit().size(); i++){
            if (s.getUnit().get(i).getEquipment().size() != t.getUnit().get(i).getEquipment().size()){
                return false;
            }
        }
        //At this point, the squads, number of units in squads and the equipment sizes are the same. Now check the actual equipment.
        for (int i = 0; i < s.getUnit().size(); i++){
            for (int j = 0; j < s.getUnit().get(i).getEquipment().size(); j++){
                if (!s.getUnit().get(i).getEquipment().get(j).getName().equals(t.getUnit().get(i).getEquipment().get(j).getName())
                        && !(s.getUnit().get(i).getEquipment().get(j).getCount() == t.getUnit().get(i).getEquipment().get(j).getCount())){
                    return false;
                }
            }
        }
        //All of the equipment, squad counts and names are the same, therefore the squads are identical, return true.
        return true;
    }
}

