import com.mongodb.client.MongoCollection;
import com.mongodb.client.MongoCursor;
import com.mongodb.client.MongoDatabase;
import org.bson.Document;

import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;

/*
    Main class for the metagame maintenance.
 */

public class MetagameMaintenance {

    public static void runMaintenance(MongoDatabase db){

        List<List<Integer>> solOptions = new ArrayList<>();
        List<List<String>> solParameters = new ArrayList<>();
        List<List<Integer>> proOptions = new ArrayList<>();
        List<List<String>> proParameters = new ArrayList<>();

        //Creates Armies From randomly selected squads.
        Army solutionArmy = createArmy("Solution Army", solOptions, solParameters);
        Army problemArmy = createArmy("Problem Army", proOptions, proParameters);

        int tempPsy = 0;

        /*
            Now we have two armies on hand, we print both of them out and play a game with them.
         */

        System.out.println();
        System.out.println("The name of each squad for the first army, its options, parameters and the ratings are presented below.");
        System.out.println();
        for (int i = 0; i < solutionArmy.getSquad().size(); i++){
            System.out.println("Squad named: " + solutionArmy.getSquad().get(i).getName());
            System.out.println("It has the following options applied: " + solOptions.get(i) + " with the following parameters " + solParameters.get(i) + ",");
            for (int j = 0; j < solutionArmy.getSquad().get(i).getUnit().size(); j++){
                System.out.println("It has the following Unit " + solutionArmy.getSquad().get(i).getUnit().get(j).getName() + " with the unit count equal to " + solutionArmy.getSquad().get(i).getUnit().get(j).getCount() + ".");
                for (int k = 0; k < solutionArmy.getSquad().get(i).getUnit().get(j).getEquipment().size(); k++){
                    System.out.println("It has the following Equipment " + solutionArmy.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getName() + " with the count equal to " + solutionArmy.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getCount() + ".");
                }
            }
            System.out.println();
        }

        System.out.println();
        System.out.println("The name of each squad for the second army, its options, parameters and the ratings are presented below.");
        System.out.println();
        for (int i = 0; i < problemArmy.getSquad().size(); i++){
            System.out.println("Squad named: " + problemArmy.getSquad().get(i).getName());
            System.out.println("It has the following options applied: " + proOptions.get(i) + " with the following parameters " + proParameters.get(i) + ".");
            for (int j = 0; j < problemArmy.getSquad().get(i).getUnit().size(); j++){
                System.out.println("It has the following Unit " + problemArmy.getSquad().get(i).getUnit().get(j).getName() + " with the unit count equal to " + problemArmy.getSquad().get(i).getUnit().get(j).getCount() + ".");
                for (int k = 0; k < problemArmy.getSquad().get(i).getUnit().get(j).getEquipment().size(); k++){
                    System.out.println("It has the following Equipment " + problemArmy.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getName() + " with the count equal to " + problemArmy.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getCount() + ".");
                }
            }
            System.out.println();
        }


        /*
            We now enter a process similar to revision, to calculate the expected and gathered results.
            Those results are then retained in a similar manner to retention.
         */

        double pointRatio = solutionArmy.getCost()/200.0 - problemArmy.getCost()/200.0;

        //Calculate values for methods
        double[] calculated = Retrieve.armyCalculation(solutionArmy);
        double savgToughness = calculated[0];
        double savgWToughness = calculated[1];
        double savgAS = calculated[2];
        double savgWS = calculated[3];
        double[] calculated2 = Retrieve.armyCalculation(problemArmy);
        double pavgToughness = calculated2[0];
        double pavgWToughness = calculated2[1];
        double pavgAS = calculated2[2];
        double pavgWS = calculated2[3];

        //Calculations for the solutionArmy
        double sratio = Retrieve.ratioOfSquads(solutionArmy,problemArmy);
        double movement = Retrieve.movementRate(solutionArmy, problemArmy)*CBRMain.movementWeight;
        double psychic = Retrieve.psychicRate(solutionArmy, problemArmy)*CBRMain.psychicWeight;
        double shooting = Retrieve.shootingRate(solutionArmy, problemArmy, pavgToughness, savgToughness, pavgAS, savgAS)*CBRMain.shootingWeight;
        double assault = Retrieve.assaultRate(solutionArmy, problemArmy, pavgWToughness, savgWToughness, pavgAS, savgAS, pavgWS, savgWS)*CBRMain.assaultWeight;
        double aExpected = (sratio + movement + psychic + shooting + assault) / (CBRMain.movementWeight+CBRMain.psychicWeight+CBRMain.shootingWeight+CBRMain.assaultWeight+1);
        aExpected = aExpected + pointRatio*0.5;


        //Test print outs:
        System.out.println("Squad Ratio = " +sratio);
        System.out.println("Movement (Solution Army) = " + movement);
        System.out.println("Psychic (Solution Army) = " + psychic);
        System.out.println("Shooting (Solution Army) = " + shooting);
        System.out.println("Assault (Solution Army) = " + assault);
        System.out.println("Expected (Solution Army) = " + aExpected);

        String s;
        Scanner input = new Scanner(System.in);
        double k = 10.0;

        System.out.println("Please input all the wins, losses and draws by each squad of the First Army.");
        for (int i = 0; i < solutionArmy.getSquad().size(); i++){
            System.out.println("Please type in w, W, or win for wins, l, L or loss for loss and d, D or draw for draw for the squad referred to as " + solutionArmy.getSquad().get(i).getName() + ".");
            s = input.nextLine();
            double score = 0;
            //Each score will get it's own separate calculation for rating.
            switch (s) {
                case "l":
                case "L":
                case "loss":
                    //If a squad looses, it will lose rating based on the point ratio and the strength ratio. The other ratings will be adjusted in integers if it had advantage or not.
                    score = 0;

                    if (movement >= CBRMain.movementWeight / 2) {
                        CBRMain.move--;
                    } else {
                        CBRMain.move++;
                    }
                    if (psychic >= CBRMain.psychicWeight / 2) {
                        tempPsy--;
                    } else {
                        tempPsy++;
                    }
                    if (shooting >= CBRMain.shootingWeight / 2) {
                        CBRMain.shoot--;
                    } else {
                        CBRMain.shoot++;
                    }
                    if (assault >= CBRMain.assaultWeight / 2) {
                        CBRMain.melee--;
                    } else {
                        CBRMain.melee++;
                    }
                    break;
                case "d":
                case "D":
                case "draw":
                    //If a squad ties, it will lose or gain rating based on the point ratio and the strength ratio. The other ratings will be adjusted in integers if it had advantage or not.
                    score = 0.5;

                    if (movement > CBRMain.movementWeight / 2 - 0.2 * CBRMain.movementWeight && CBRMain.movementWeight / 2 + 0.2 * CBRMain.movementWeight < 0.6) {
                        CBRMain.move++;
                    } else {
                        CBRMain.move--;
                    }
                    if (psychic > CBRMain.psychicWeight / 2 - 0.2 * CBRMain.psychicWeight && CBRMain.psychicWeight / 2 + 0.2 * CBRMain.psychicWeight < 0.6) {
                        tempPsy++;
                    } else {
                        tempPsy--;
                    }
                    if (shooting > CBRMain.shootingWeight / 2 - 0.2 * CBRMain.shootingWeight && CBRMain.shootingWeight / 2 + 0.2 * CBRMain.shootingWeight < 0.6) {
                        CBRMain.shoot++;
                    } else {
                        CBRMain.shoot--;
                    }
                    if (assault > CBRMain.assaultWeight / 2 - 0.2 * CBRMain.assaultWeight && CBRMain.assaultWeight / 2 + 0.2 * CBRMain.assaultWeight < 0.6) {
                        CBRMain.melee++;
                    } else {
                        CBRMain.melee--;
                    }
                    break;
                case "w":
                case "W":
                case "win":
                    //If a squad wins, it will gain rating based on the point ratio and the strength ratio. The other ratings will be adjusted in integers if it had advantage or not.
                    score = 1;

                    if (movement >= CBRMain.movementWeight / 2) {
                        CBRMain.move++;
                    } else {
                        CBRMain.move--;
                    }
                    if (psychic >= CBRMain.psychicWeight / 2) {
                        tempPsy++;
                    } else {
                        tempPsy--;
                    }
                    if (shooting >= CBRMain.shootingWeight / 2) {
                        CBRMain.shoot++;
                    } else {
                        CBRMain.shoot--;
                    }
                    if (assault >= CBRMain.assaultWeight / 2) {
                        CBRMain.melee++;
                    } else {
                        CBRMain.melee--;
                    }
                    break;
            }
            solutionArmy.getSquad().get(i).setRating(solutionArmy.getSquad().get(i).getRating() + k*(score - aExpected));
        }

        //Calculations for the problemArmy
        sratio = Retrieve.ratioOfSquads(problemArmy, solutionArmy);
        movement = Retrieve.movementRate(problemArmy, solutionArmy)*CBRMain.movementWeight;
        psychic = Retrieve.psychicRate(problemArmy, solutionArmy)*CBRMain.psychicWeight;
        shooting = Retrieve.shootingRate(problemArmy, solutionArmy, savgToughness, pavgToughness, savgAS, pavgAS)*CBRMain.shootingWeight;
        assault = Retrieve.assaultRate(problemArmy, solutionArmy, savgWToughness, pavgWToughness, savgAS, pavgAS, savgWS, pavgWS)*CBRMain.assaultWeight;
        double pExpected = (sratio + movement + psychic + shooting + assault) / (CBRMain.movementWeight+CBRMain.psychicWeight+CBRMain.shootingWeight+CBRMain.assaultWeight+1);
        pExpected = pExpected - pointRatio*0.5;

        System.out.println("Movement (Problem Army) = " + movement);
        System.out.println("Psychic (Problem Army) = " + psychic);
        System.out.println("Shooting (Problem Army) = " + shooting);
        System.out.println("Assault (Problem Army) = " + assault);
        System.out.println("Expected (Problem Army) = " + pExpected);

        System.out.println();
        System.out.println("Input finished for the First army, please input wins, draws and losses for the Second army.");
        for (int i = 0; i < problemArmy.getSquad().size(); i++){
            System.out.println("Please type in w, W, or win for wins, l, L or loss for loss and d, D or draw for draw for the squad referred to as " + problemArmy.getSquad().get(i).getName() + ".");
            s = input.nextLine();
            double score = 0;
            //Each score will get it's own separate calculation for rating.
            switch (s) {
                case "l":
                case "L":
                case "loss":
                    //If a squad looses, it will lose rating based on the point ratio and the strength ratio. The other ratings will be adjusted in integers if it had advantage or not.
                    score = 0;

                    if (movement >= CBRMain.movementWeight / 2) {
                        CBRMain.move--;
                    } else {
                        CBRMain.move++;
                    }
                    if (psychic >= CBRMain.psychicWeight / 2) {
                        tempPsy--;
                    } else {
                        tempPsy++;
                    }
                    if (shooting >= CBRMain.shootingWeight / 2) {
                        CBRMain.shoot--;
                    } else {
                        CBRMain.shoot++;
                    }
                    if (assault >= CBRMain.assaultWeight / 2) {
                        CBRMain.melee--;
                    } else {
                        CBRMain.melee++;
                    }
                    break;
                case "d":
                case "D":
                case "draw":
                    //If a squad ties, it will lose or gain rating based on the point ratio and the strength ratio. The other ratings will be adjusted in integers if it had advantage or not.
                    score = 0.5;

                    if (movement > CBRMain.movementWeight / 2 - 0.2 * CBRMain.movementWeight && CBRMain.movementWeight / 2 + 0.2 * CBRMain.movementWeight < 0.6) {
                        CBRMain.move++;
                    } else {
                        CBRMain.move--;
                    }
                    if (psychic > CBRMain.psychicWeight / 2 - 0.2 * CBRMain.psychicWeight && CBRMain.psychicWeight / 2 + 0.2 * CBRMain.psychicWeight < 0.6) {
                        tempPsy++;
                    } else {
                        tempPsy--;
                    }
                    if (shooting > CBRMain.shootingWeight / 2 - 0.2 * CBRMain.shootingWeight && CBRMain.shootingWeight / 2 + 0.2 * CBRMain.shootingWeight < 0.6) {
                        CBRMain.shoot++;
                    } else {
                        CBRMain.shoot--;
                    }
                    if (assault > CBRMain.assaultWeight / 2 - 0.2 * CBRMain.assaultWeight && CBRMain.assaultWeight / 2 + 0.2 * CBRMain.assaultWeight < 0.6) {
                        CBRMain.melee++;
                    } else {
                        CBRMain.melee--;
                    }
                    break;
                case "w":
                case "W":
                case "win":
                    //If a squad wins, it will gain rating based on the point ratio and the strength ratio. The other ratings will be adjusted in integers if it had advantage or not.
                    score = 1;

                    if (movement >= CBRMain.movementWeight / 2) {
                        CBRMain.move++;
                    } else {
                        CBRMain.move--;
                    }
                    if (psychic >= CBRMain.psychicWeight / 2) {
                        tempPsy++;
                    } else {
                        tempPsy--;
                    }
                    if (shooting >= CBRMain.shootingWeight / 2) {
                        CBRMain.shoot++;
                    } else {
                        CBRMain.shoot--;
                    }
                    if (assault >= CBRMain.assaultWeight / 2) {
                        CBRMain.melee++;
                    } else {
                        CBRMain.melee--;
                    }
                    break;
            }
            problemArmy.getSquad().get(i).setRating(problemArmy.getSquad().get(i).getRating() + k*(score - pExpected));
        }

        /*
            Psykers are somewhat special, and we do not want to demote the psychic weight when there are no psykers in battle.
            Therefore, we count up all the psykers in a battle, and if there were none we simply reset the psychic weight measurement to 0.
         */

        int countPsykers = 0;
        for (int i = 0; i < solutionArmy.getSquad().size(); i++){
            for (int j = 0; j < solutionArmy.getSquad().get(i).getUnit().size(); j++){
                if (solutionArmy.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    for (int l = 0; l < solutionArmy.getSquad().get(i).getUnit().get(j).getRules().size(); l++) {
                        if (solutionArmy.getSquad().get(i).getUnit().get(j).getRules().get(l).contains("Psyker") ||
                                solutionArmy.getSquad().get(i).getUnit().get(j).getRules().get(l).contains("Psychic Pilot") ||
                                solutionArmy.getSquad().get(i).getUnit().get(j).getRules().get(l).contains("Brotherhood of Psykers") ||
                                solutionArmy.getSquad().get(i).getUnit().get(j).getRules().get(l).contains("Brotherhood of Sorcerers")) {
                            countPsykers++;
                        }
                    }
                }
            }
        }

        for (int i = 0; i < problemArmy.getSquad().size(); i++){
            for (int j = 0; j < problemArmy.getSquad().get(i).getUnit().size(); j++){
                if (problemArmy.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    for (int l = 0; l < problemArmy.getSquad().get(i).getUnit().get(j).getRules().size(); l++) {
                        if (problemArmy.getSquad().get(i).getUnit().get(j).getRules().get(l).contains("Psyker") ||
                                problemArmy.getSquad().get(i).getUnit().get(j).getRules().get(l).contains("Psychic Pilot") ||
                                problemArmy.getSquad().get(i).getUnit().get(j).getRules().get(l).contains("Brotherhood of Psykers") ||
                                problemArmy.getSquad().get(i).getUnit().get(j).getRules().get(l).contains("Brotherhood of Sorcerers")) {
                            countPsykers++;
                        }
                    }
                }
            }
        }

        if (countPsykers != 0){
            CBRMain.psy += tempPsy;
        }

        //Still tests!
        System.out.println("Prediction for movement = " + CBRMain.move);
        System.out.println("Prediction for psychic = " + CBRMain.psy);
        System.out.println("Prediction for shooting = " + CBRMain.shoot);
        System.out.println("Prediction for melee = " + CBRMain.melee);

        //Run retention and retain squads (not armies!) individually, and their ratings (if basic).

        updateSquads(solutionArmy, solOptions, solParameters);
        updateSquads(problemArmy, proOptions, proParameters);

        for (int i = 0; i < solutionArmy.getSquad().size(); i++) {

            int checkDuplicates = 0;
            double rating = 0;
            for (int j = 0; j < solutionArmy.getSquad().size(); j++) {
                if (CaseBaseLoading.checkDuplicate(solutionArmy.getSquad().get(i), solutionArmy.getSquad().get(j))) {
                    checkDuplicates++;
                    rating += solutionArmy.getSquad().get(j).getRating();
                }
            }

            for (int j = 0; j < problemArmy.getSquad().size(); j++) {
                if (CaseBaseLoading.checkDuplicate(solutionArmy.getSquad().get(i), problemArmy.getSquad().get(j))) {
                    checkDuplicates++;
                    rating += problemArmy.getSquad().get(j).getRating();
                }
            }

            for (int j = 0; j < CBRMain.SquadList.size(); j++) {
                if (CBRMain.SquadList.get(j).getName().contains(solutionArmy.getSquad().get(i).getName())) {
                    if (CaseBaseLoading.checkDuplicate(CBRMain.SquadList.get(j), solutionArmy.getSquad().get(i))) {
                        CBRMain.SquadList.get(j).setRating(rating / checkDuplicates);
                    }
                }
            }

            System.out.println(solutionArmy.getSquad().get(i).getName() + " " + solutionArmy.getSquad().get(i).getRating());
        }

        for (int i = 0; i < problemArmy.getSquad().size(); i++) {

            int checkDuplicates = 0;
            double rating = 0;
            for (int j = 0; j < solutionArmy.getSquad().size(); j++) {
                if (CaseBaseLoading.checkDuplicate(problemArmy.getSquad().get(i), solutionArmy.getSquad().get(j))) {
                    checkDuplicates++;
                    rating += solutionArmy.getSquad().get(j).getRating();
                }
            }

            for (int j = 0; j < problemArmy.getSquad().size(); j++) {
                if (CaseBaseLoading.checkDuplicate(problemArmy.getSquad().get(i), problemArmy.getSquad().get(j))) {
                    checkDuplicates++;
                    rating += problemArmy.getSquad().get(j).getRating();
                }
            }

            for (int j = 0; j < CBRMain.SquadList.size(); j++) {
                if (CBRMain.SquadList.get(j).getName().contains(problemArmy.getSquad().get(i).getName())) {
                    if (CaseBaseLoading.checkDuplicate(CBRMain.SquadList.get(j), problemArmy.getSquad().get(i))) {
                        CBRMain.SquadList.get(j).setRating(rating / checkDuplicates);
                    }
                }
            }

            System.out.println(problemArmy.getSquad().get(i).getName() + " " + problemArmy.getSquad().get(i).getRating());
        }

        saveRatings(db);
    }

    public static Army createArmy(String name, List<List<Integer>> options, List<List<String>> parameters){

        double cost = 0;
        double rating = 0;
        List<Squad> squadList = new ArrayList<>();
        int iterations = 0;

        while (cost < 200 && iterations < 20) {

            int random = randomWithRange(0, CBRMain.SquadList.size()-1);

            if (cost+CBRMain.SquadList.get(random).getCost() <= 200){
                squadList.add(new Squad(CBRMain.SquadList.get(random)));
                options.add(CBRMain.squadOptions.get(random));
                parameters.add(CBRMain.squadParameters.get(random));
                cost += CBRMain.SquadList.get(random).getCost();
                rating += CBRMain.SquadList.get(random).getRating();
            }

            iterations++;
        }

        /*
            Try to add options to fine-tune the squad.
            We will run five (this number can be changed if desired) iterations on each unit in each squad to try to add a random option.
            This ensures an equal distribution of randomness through the creation process.
         */

        int end = 0;
        while(end < squadList.size() * 5 && cost <= 200) {
            for (int j = 0; j < squadList.size(); j++) {
                for (int i = 0; i < squadList.get(j).getUnit().size(); i++) {
                    //Picks an option at random from a unit.
                    int max = squadList.get(j).getUnit().get(i).getOptions().size() - 1;
                    int random = randomWithRange(0, max);

                    //Do logic to try to use options.
                    Squad x = new Squad(squadList.get(j));
                    GeneralKnowledge.Option(x.getUnit().get(i).getOptions().get(random), "None", x);

                    double tempCost = cost - squadList.get(j).getCost() + x.getCost();
                    //See if we have a legal costing army.
                    if (tempCost <= 200 && !CaseBaseLoading.checkDuplicate(x, squadList.get(j))) {
                        squadList.set(j, x);
                        options.get(j).add(x.getUnit().get(i).getOptions().get(random));
                        parameters.get(j).add("None");
                        cost = tempCost;
                    }
                }
            }

            end++;
        }

        rating = rating/squadList.size();

        return new Army(name, cost, rating, squadList);
    }

    public static void saveRatings(MongoDatabase db){

        for (int i = 0; i < CBRMain.SquadList.size(); i++){

            boolean updated = false;

            MongoCollection<Document> squadCollection = db.getCollection("unit");
            MongoCursor<Document> cursor = squadCollection.find().iterator();
            Document doc;

            //First we check if the squad exists in the case base. If it does we just update it's rating variable.
            for (int j = 0; j < squadCollection.count(); j++){
                doc = cursor.next();
                if (CBRMain.SquadList.get(i).getName().equals(doc.getString("Name"))){
                    squadCollection.updateOne(new Document("Name", CBRMain.SquadList.get(i).getName()),
                            new Document("$set", new Document("Rating", CBRMain.SquadList.get(i).getRating())));
                    updated = true;
                }
            }

            //The squad does not exist in the case base. Insert it.
            if (!updated){
                //Adds the units to the case base.
                List<Document> unit = new ArrayList<>();

                for (int j = 0; j < CBRMain.SquadList.get(i).getUnit().size(); j++){

                    List<Document> o = new ArrayList<>();
                    for (int k = 0; k < CBRMain.SquadList.get(i).getUnit().get(j).getOptions().size(); k++){
                        o.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getOptions().get(k)));
                    }

                    List<Document> e = new ArrayList<>();
                    for (int k = 0; k < CBRMain.SquadList.get(i).getUnit().get(j).getEquipment().size(); k++){
                        e.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getEquipment().get(k).getName())
                                .append("Count", CBRMain.SquadList.get(i).getUnit().get(j).getEquipment().get(k).getCount()));
                    }

                    List<Document> r = new ArrayList<>();
                    for (int k = 0; k < CBRMain.SquadList.get(i).getUnit().get(j).getRules().size(); k++){
                        r.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getRules().get(k)));
                    }

                    //add Infantry, Bikes and Artillery.
                    if (CBRMain.SquadList.get(i).getUnit().get(j).getType().contains("Infantry")
                            || CBRMain.SquadList.get(i).getUnit().get(j).getType().contains("Bike")
                            || CBRMain.SquadList.get(i).getUnit().get(j).getType().contains("Artillery")){

                        unit.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getName())
                                .append("Weapon Skill", CBRMain.SquadList.get(i).getUnit().get(j).getWS())
                                .append("Ballistics Skill", CBRMain.SquadList.get(i).getUnit().get(j).getBS())
                                .append("Strength", CBRMain.SquadList.get(i).getUnit().get(j).getS())
                                .append("Toughness", CBRMain.SquadList.get(i).getUnit().get(j).getT())
                                .append("Wounds", CBRMain.SquadList.get(i).getUnit().get(j).getW())
                                .append("Initiative", CBRMain.SquadList.get(i).getUnit().get(j).getI())
                                .append("Attacks", CBRMain.SquadList.get(i).getUnit().get(j).getA())
                                .append("Leadership", CBRMain.SquadList.get(i).getUnit().get(j).getLD())
                                .append("Armour Save", CBRMain.SquadList.get(i).getUnit().get(j).getAS())
                                .append("Unit Type", CBRMain.SquadList.get(i).getUnit().get(j).getType())
                                .append("Unique", CBRMain.SquadList.get(i).getUnit().get(j).isUnique())
                                .append("Count", CBRMain.SquadList.get(i).getUnit().get(j).getCount())
                                .append("Faction", CBRMain.SquadList.get(i).getUnit().get(j).getFaction())
                                .append("Equipment", e)
                                .append("Special Rules", r)
                                .append("Options", o)
                                .append("Warlord Trait", CBRMain.SquadList.get(i).getUnit().get(j).getTrait()));
                    }

                    //add Walkers
                    else if (CBRMain.SquadList.get(i).getUnit().get(j).getType().equals("Vehicle-Walker")){

                        unit.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getName())
                                .append("Weapon Skill", CBRMain.SquadList.get(i).getUnit().get(j).getWS())
                                .append("Ballistics Skill", CBRMain.SquadList.get(i).getUnit().get(j).getBS())
                                .append("Strength", CBRMain.SquadList.get(i).getUnit().get(j).getS())
                                .append("F", CBRMain.SquadList.get(i).getUnit().get(j).getFront())
                                .append("S", CBRMain.SquadList.get(i).getUnit().get(j).getSide())
                                .append("R", CBRMain.SquadList.get(i).getUnit().get(j).getRear())
                                .append("Initiative", CBRMain.SquadList.get(i).getUnit().get(j).getI())
                                .append("Attacks", CBRMain.SquadList.get(i).getUnit().get(j).getA())
                                .append("HP", CBRMain.SquadList.get(i).getUnit().get(j).getHp())
                                .append("Unit Type", CBRMain.SquadList.get(i).getUnit().get(j).getType())
                                .append("Unique", CBRMain.SquadList.get(i).getUnit().get(j).isUnique())
                                .append("Count", CBRMain.SquadList.get(i).getUnit().get(j).getCount())
                                .append("Faction", CBRMain.SquadList.get(i).getUnit().get(j).getFaction())
                                .append("Equipment", e)
                                .append("Special Rules", r)
                                .append("Options", o)
                                .append("Warlord Trait", CBRMain.SquadList.get(i).getUnit().get(j).getTrait()));
                    }

                    //add everything else
                    else{

                        unit.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getName())
                                .append("Ballistics Skill", CBRMain.SquadList.get(i).getUnit().get(j).getBS())
                                .append("Strength", CBRMain.SquadList.get(i).getUnit().get(j).getS())
                                .append("F", CBRMain.SquadList.get(i).getUnit().get(j).getFront())
                                .append("S", CBRMain.SquadList.get(i).getUnit().get(j).getSide())
                                .append("R", CBRMain.SquadList.get(i).getUnit().get(j).getRear())
                                .append("HP", CBRMain.SquadList.get(i).getUnit().get(j).getHp())
                                .append("Unit Type", CBRMain.SquadList.get(i).getUnit().get(j).getType())
                                .append("Unique", CBRMain.SquadList.get(i).getUnit().get(j).isUnique())
                                .append("Count", CBRMain.SquadList.get(i).getUnit().get(j).getCount())
                                .append("Faction", CBRMain.SquadList.get(i).getUnit().get(j).getFaction())
                                .append("Equipment", e)
                                .append("Special Rules", r)
                                .append("Options", o)
                                .append("Warlord Trait", CBRMain.SquadList.get(i).getUnit().get(j).getTrait()));
                    }
                }

                //Adds the options and parameters lists to the case base.
                List<Document> o = new ArrayList<>();
                List<Document> par = new ArrayList<>();
                for (int j = 0; j < CBRMain.squadOptions.get(i).size(); j++){
                    o.add(new Document().append("Name", CBRMain.squadOptions.get(i).get(j)));
                    par.add(new Document().append("Name", CBRMain.squadParameters.get(i).get(j)));
                }

                //Insert the squad here. So far, units, options and parameters are not saved correctly.
                squadCollection.insertOne(new Document().append("Name", CBRMain.SquadList.get(i).getName())
                        .append("Units", unit)
                        .append("Cost", CBRMain.SquadList.get(i).getCost())
                        .append("Role", CBRMain.SquadList.get(i).getRole())
                        .append("Base", CBRMain.SquadList.get(i).getBase())
                        .append("Options", o)
                        .append("Parameters", par)
                        .append("Rating", CBRMain.SquadList.get(i).getRating()));

            }
        }
    }

    public static void updateSquads(Army a, List<List<Integer>> opt, List<List<String>> par){

        for (int i = 0; i < a.getSquad().size(); i++) {

            int countOccurences = 1;
            for (int j = 0; j < CBRMain.SquadList.size(); j++) {
                if (CBRMain.SquadList.get(j).getName().contains(a.getSquad().get(i).getName())) {
                    countOccurences++;
                }
            }

            int countDuplicates = 0;

            //Make insert a new squad (so that the naming is preserved for the army).
            Squad insert = new Squad(a.getSquad().get(i));

            //Set the name and the base of the insert to be different, so that it is a different case.
            insert.setName(a.getSquad().get(i).getName() + " " + countOccurences);
            insert.setBase(false);

            for (int l = 0; l < CBRMain.SquadList.size(); l++) {
                if (CaseBaseLoading.checkDuplicate(CBRMain.SquadList.get(l), insert)) {
                    countDuplicates++;
                }
            }

            if (countDuplicates == 0) {
                CBRMain.SquadList.add(insert);
                CBRMain.squadOptions.add(opt.get(i));
                CBRMain.squadParameters.add(par.get(i));
            }
        }
    }

    public static int randomWithRange(int min, int max){
        int range = (max - min) + 1;
        return (int)(Math.random() * range) + min;
    }
}
