import com.mongodb.client.MongoCollection;
import com.mongodb.client.MongoCursor;
import com.mongodb.client.MongoDatabase;
import org.bson.Document;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/*
    Retains the solutions into the case base. This involves using the case base (e.g the database).
 */
public class Retain {

    public static void mainRetain(Army a, Army p, MongoDatabase db) throws ParseException {

        /*
            Retains the solution army (if it has changed from the original). Also retains the rating for the problem army.
        */

        int duplicateArmies = 0;

        MongoCollection<Document> armyCollection = db.getCollection("army");

        int duplicateArmyNumber = -1;
        for (int i = 0; i < CBRMain.ArmyList.size(); i++){
            if (checkArmies(a, CBRMain.ArmyList.get(i), i)){
                duplicateArmies++;
                duplicateArmyNumber = i;
            }
        }

        if (duplicateArmies == 0) {
            addArmy(a, db);
        }
        else{
            String s = String.valueOf(duplicateArmyNumber);
            //The rating for the solution army is instead updated.
            armyCollection.updateOne(new Document("ID", s), new Document("$set", new Document("Rating", a.getRating())));
        }

        //The rating for the problem army is updated.
        armyCollection.updateOne(new Document("Name", p.getName()), new Document("$set", new Document("Rating", p.getRating())));

        if (duplicateArmies != 0) {
            CBRMain.retainExplanations.add("Army ratings for both armies have been updated.");
        }
        else{
            CBRMain.retainExplanations.add("The Army Rating for the problem army has been updated. The solution army has been added to the case base.");
        }
        CBRMain.retainExplanations.add("The Squad Ratings for all squads have been updated.");
        CBRMain.retainExplanations.add("");
        CBRMain.retainExplanations.add("The following are the new armies, their ratings, and the ratings of each individual squad.");
        CBRMain.retainExplanations.add("");
        CBRMain.retainExplanations.add("Solution Army Rating: " + a.getRating() + ".");
        CBRMain.retainExplanations.add("");

        for (int i = 0; i < a.getSquad().size(); i++){
            boolean duplicate = false;
            for (int j = 0; j < CBRMain.SquadList.size(); j++) {
                if (CBRMain.SquadList.get(j).getName().contains(a.getSquad().get(i).getName())) {
                    //Check if there is a single duplicate in the system. If there is, we do not add a squad.
                    if (CaseBaseLoading.checkDuplicate(CBRMain.SquadList.get(j), a.getSquad().get(i))) {
                        CBRMain.retainExplanations.add("For the squad named " + a.getSquad().get(i).getName() + " the rating is now " + CBRMain.SquadList.get(j).getRating() + ".");
                        duplicate = true;
                    }
                }
            }
            if (!duplicate){
                CBRMain.retainExplanations.add("A new squad has been added to the army and knowledge base named " + a.getSquad().get(i).getName() + " with the rating " + a.getSquad().get(i).getRating() + ".");
            }
        }
        
        CBRMain.retainExplanations.add("");
        CBRMain.retainExplanations.add("Problem Army Rating: " + p.getRating());
        CBRMain.retainExplanations.add("");
        for (int i = 0; i < p.getSquad().size(); i++){
            boolean duplicate = false;
            for (int j = 0; j < CBRMain.SquadList.size(); j++) {
                if (CBRMain.SquadList.get(j).getName().contains(p.getSquad().get(i).getName())) {
                    //Check if there is a single duplicate in the system. If there is, we do not add a squad.
                    if (CaseBaseLoading.checkDuplicate(CBRMain.SquadList.get(j), p.getSquad().get(i))) {
                        CBRMain.retainExplanations.add("For the squad named " + p.getSquad().get(i).getName() + " the rating is now " + CBRMain.SquadList.get(j).getRating() + ".");
                        duplicate = true;
                    }
                }
            }
            if (!duplicate){
                CBRMain.retainExplanations.add("A new squad has been added to the army and knowledge base named " + p.getSquad().get(i).getName() + " with the rating " + p.getSquad().get(i).getRating() + ".");
            }
        }

        /*
            Retains all the squads from the SquadList.
        */

        for (int i = 0; i < CBRMain.SquadList.size(); i++){

            boolean updated = false;

            MongoCollection<Document> squadCollection = db.getCollection("unit");
            MongoCursor<Document> cursor = squadCollection.find().iterator();
            Document doc;

            //First we check if the squad exists in the case base. If it does we just update it's rating variable.
            for (int j = 0; j < squadCollection.count(); j++){
                doc = cursor.next();
                if (CBRMain.SquadList.get(i).getName().equals(doc.getString("Name"))){
                    squadCollection.updateOne(new Document("Name", CBRMain.SquadList.get(i).getName()),
                            new Document("$set", new Document("Rating", CBRMain.SquadList.get(i).getRating())));
                    updated = true;
                }
            }

            //The squad does not exist in the case base. Insert it.
            if (!updated){
                //Adds the units to the case base.
                List<Document> unit = new ArrayList<>();

                for (int j = 0; j < CBRMain.SquadList.get(i).getUnit().size(); j++){

                    List<Document> o = new ArrayList<>();
                    for (int k = 0; k < CBRMain.SquadList.get(i).getUnit().get(j).getOptions().size(); k++){
                        o.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getOptions().get(k)));
                    }

                    List<Document> e = new ArrayList<>();
                    for (int k = 0; k < CBRMain.SquadList.get(i).getUnit().get(j).getEquipment().size(); k++){
                        e.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getEquipment().get(k).getName())
                                            .append("Count", CBRMain.SquadList.get(i).getUnit().get(j).getEquipment().get(k).getCount()));
                    }

                    List<Document> r = new ArrayList<>();
                    for (int k = 0; k < CBRMain.SquadList.get(i).getUnit().get(j).getRules().size(); k++){
                        r.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getRules().get(k)));
                    }

                    //add Infantry, Bikes and Artillery.
                    if (CBRMain.SquadList.get(i).getUnit().get(j).getType().contains("Infantry")
                        || CBRMain.SquadList.get(i).getUnit().get(j).getType().contains("Bike")
                        || CBRMain.SquadList.get(i).getUnit().get(j).getType().contains("Jump")
                        || CBRMain.SquadList.get(i).getUnit().get(j).getType().contains("Artillery")){

                        unit.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getName())
                                .append("Weapon Skill", CBRMain.SquadList.get(i).getUnit().get(j).getWS())
                                .append("Ballistics Skill", CBRMain.SquadList.get(i).getUnit().get(j).getBS())
                                .append("Strength", CBRMain.SquadList.get(i).getUnit().get(j).getS())
                                .append("Toughness", CBRMain.SquadList.get(i).getUnit().get(j).getT())
                                .append("Wounds", CBRMain.SquadList.get(i).getUnit().get(j).getW())
                                .append("Initiative", CBRMain.SquadList.get(i).getUnit().get(j).getI())
                                .append("Attacks", CBRMain.SquadList.get(i).getUnit().get(j).getA())
                                .append("Leadership", CBRMain.SquadList.get(i).getUnit().get(j).getLD())
                                .append("Armour Save", CBRMain.SquadList.get(i).getUnit().get(j).getAS())
                                .append("Unit Type", CBRMain.SquadList.get(i).getUnit().get(j).getType())
                                .append("Unique", CBRMain.SquadList.get(i).getUnit().get(j).isUnique())
                                .append("Count", CBRMain.SquadList.get(i).getUnit().get(j).getCount())
                                .append("Faction", CBRMain.SquadList.get(i).getUnit().get(j).getFaction())
                                .append("Equipment", e)
                                .append("Special Rules", r)
                                .append("Options", o)
                                .append("Warlord Trait", CBRMain.SquadList.get(i).getUnit().get(j).getTrait()));
                    }

                    //add Walkers
                    else if (CBRMain.SquadList.get(i).getUnit().get(j).getType().equals("Vehicle-Walker")){

                        unit.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getName())
                                .append("Weapon Skill", CBRMain.SquadList.get(i).getUnit().get(j).getWS())
                                .append("Ballistics Skill", CBRMain.SquadList.get(i).getUnit().get(j).getBS())
                                .append("Strength", CBRMain.SquadList.get(i).getUnit().get(j).getS())
                                .append("F", CBRMain.SquadList.get(i).getUnit().get(j).getFront())
                                .append("S", CBRMain.SquadList.get(i).getUnit().get(j).getSide())
                                .append("R", CBRMain.SquadList.get(i).getUnit().get(j).getRear())
                                .append("Initiative", CBRMain.SquadList.get(i).getUnit().get(j).getI())
                                .append("Attacks", CBRMain.SquadList.get(i).getUnit().get(j).getA())
                                .append("HP", CBRMain.SquadList.get(i).getUnit().get(j).getHp())
                                .append("Unit Type", CBRMain.SquadList.get(i).getUnit().get(j).getType())
                                .append("Unique", CBRMain.SquadList.get(i).getUnit().get(j).isUnique())
                                .append("Count", CBRMain.SquadList.get(i).getUnit().get(j).getCount())
                                .append("Faction", CBRMain.SquadList.get(i).getUnit().get(j).getFaction())
                                .append("Equipment", e)
                                .append("Special Rules", r)
                                .append("Options", o)
                                .append("Warlord Trait", CBRMain.SquadList.get(i).getUnit().get(j).getTrait()));
                    }

                    //add everything else
                    else{

                        unit.add(new Document().append("Name", CBRMain.SquadList.get(i).getUnit().get(j).getName())
                                .append("Ballistics Skill", CBRMain.SquadList.get(i).getUnit().get(j).getBS())
                                .append("Strength", CBRMain.SquadList.get(i).getUnit().get(j).getS())
                                .append("F", CBRMain.SquadList.get(i).getUnit().get(j).getFront())
                                .append("S", CBRMain.SquadList.get(i).getUnit().get(j).getSide())
                                .append("R", CBRMain.SquadList.get(i).getUnit().get(j).getRear())
                                .append("HP", CBRMain.SquadList.get(i).getUnit().get(j).getHp())
                                .append("Unit Type", CBRMain.SquadList.get(i).getUnit().get(j).getType())
                                .append("Unique", CBRMain.SquadList.get(i).getUnit().get(j).isUnique())
                                .append("Count", CBRMain.SquadList.get(i).getUnit().get(j).getCount())
                                .append("Faction", CBRMain.SquadList.get(i).getUnit().get(j).getFaction())
                                .append("Equipment", e)
                                .append("Special Rules", r)
                                .append("Options", o)
                                .append("Warlord Trait", CBRMain.SquadList.get(i).getUnit().get(j).getTrait()));
                    }
                }

                //Adds the options and parameters lists to the case base.
                List<Document> o = new ArrayList<>();
                List<Document> par = new ArrayList<>();
                for (int j = 0; j < CBRMain.squadOptions.get(i).size(); j++){
                    o.add(new Document().append("Name", CBRMain.squadOptions.get(i).get(j)));
                    par.add(new Document().append("Name", CBRMain.squadParameters.get(i).get(j)));
                }

                //Insert the squad here. So far, units, options and parameters are not saved correctly.
                squadCollection.insertOne(new Document().append("Name", CBRMain.SquadList.get(i).getName())
                                                        .append("Units", unit)
                                                        .append("Cost", CBRMain.SquadList.get(i).getCost())
                                                        .append("Role", CBRMain.SquadList.get(i).getRole())
                                                        .append("Base", CBRMain.SquadList.get(i).getBase())
                                                        .append("Options", o)
                                                        .append("Parameters", par)
                                                        .append("Rating", CBRMain.SquadList.get(i).getRating()));

            }
        }
    }

    public static void addArmy(Army a, MongoDatabase db){

        MongoCollection<Document> armyCollection = db.getCollection("army");

        //Calculates the cost of the army, to make sure that the cost is absolutely correct.
        double cost = 0;
        for (int j = 0; j < a.getSquad().size(); j++) {
            cost += a.getSquad().get(j).getCost();
        }

        //Three list are necessary to place into the army: Squad List, Options and Parameters.
        List<Document> s = new ArrayList<>();
        List<List<Document>> o = new ArrayList<>();
        List<List<Document>> par = new ArrayList<>();

        //Creates the parameters and options lists.
        for (int i = 0; i < CBRMain.options.get(CBRMain.options.size()-1).size(); i++){
            List<Document> otmp = new ArrayList<>();
            List<Document> partmp = new ArrayList<>();
            for (int j = 0; j < CBRMain.options.get(CBRMain.options.size()-1).get(i).size(); j++){
                otmp.add(new Document().append("Option", CBRMain.options.get(CBRMain.options.size()-1).get(i).get(j)));
                partmp.add(new Document().append("Name", CBRMain.parameters.get(CBRMain.parameters.size()-1).get(i).get(j)));
            }
            o.add(otmp);
            par.add(partmp);
        }

        //Creates the squad list.
        for (int i = 0; i < a.getSquad().size(); i++){
            String tmp = a.getSquad().get(i).getName();
            if(tmp.matches(".*\\d.*")){
                tmp = tmp.substring(0, tmp.lastIndexOf(" "));
            }
            s.add(new Document()
                    .append("Name", tmp)
                    .append("Options", o.get(i))
                    .append("Parameter", par.get(i)));
        }

        //Adds the army
        armyCollection.insertOne(
                new Document()
                        .append("ID", String.valueOf(CBRMain.ArmyList.size()))
                        .append("Name", a.getName() + " " + new Date(System.currentTimeMillis()))
                        .append("Squads", s)
                        .append("Cost", cost)
                        .append("Rating", a.getRating()));
    }

    public static boolean checkArmies(Army a, Army p, int x){

        //First we check if the options match.
        if (!CBRMain.options.get(x).equals(CBRMain.options.get(CBRMain.options.size()-1))){
            return false;
        }

        //Then we check if the parameters match.
        if (!CBRMain.parameters.get(x).equals(CBRMain.parameters.get(CBRMain.parameters.size()-1))){
            return false;
        }

        //If both options and parameters match (and thus we have the same amount of squads), we check if the squads match.
        List<Squad> problemList = new ArrayList<>();
        for (int k = 0; k < p.getSquad().size(); k++) {
            problemList.add(new Squad(p.getSquad().get(k)));
        }

        Army problem = new Army(p.getName(), p.getCost(), p.getRating(), problemList);

        int solutionDuplicates = 0;

        for (int i = 0; i < a.getSquad().size(); i++){
            for (int j = 0; j < problem.getSquad().size(); j++) {
                if (CaseBaseLoading.checkDuplicate(a.getSquad().get(i), problem.getSquad().get(j))) {
                    solutionDuplicates++;
                    problem.getSquad().remove(j);
                    break;
                }
            }
        }

        //If the squads match, then we have a duplicate.
        return (double) solutionDuplicates / (double) a.getSquad().size() == 1.0;
    }
}
