import java.util.ArrayList;
import java.util.List;

/*
    Adapts the retrieved army to better fit the problem army.
 */
public class Reuse {

    //Terrain is in order: Building %, Difficult Terrain %, Dangerous Terrain %, Impassable Terrain %.
    public static List<Double> terrain = Retrieve.terrain;

    //Main Reuse Method
    public static Army mainReuse(Army a, Army p) {

        /*
            First we check if the similarity is greater than 0.673, 10% higher than the NOVA 2015 tournament highest win army.
            If it is, then we stand a good chance of winning and we do not adapt the army.
         */

        double[] calculated = Retrieve.armyCalculation(a);
        double savgToughness = calculated[0];
        double savgWToughness = calculated[1];
        double savgAS = calculated[2];
        double savgWS = calculated[3];
        double[] calculated2 = Retrieve.armyCalculation(p);
        double pavgToughness = calculated2[0];
        double pavgWToughness = calculated2[1];
        double pavgAS = calculated2[2];
        double pavgWS = calculated2[3];

        double ratioOfStrength = (Retrieve.movementRate(a, p) * CBRMain.movementWeight + Retrieve.psychicRate(a, p) * CBRMain.psychicWeight + Retrieve.shootingRate(a, p, pavgToughness, savgToughness, pavgAS, savgAS)
                * CBRMain.shootingWeight + Retrieve.assaultRate(a, p, pavgWToughness, savgWToughness, pavgAS, savgAS, pavgWS, savgWS) * CBRMain.assaultWeight)
                / (CBRMain.movementWeight + CBRMain.psychicWeight + CBRMain.shootingWeight + CBRMain.assaultWeight);

        double similarity = (Retrieve.ratioOfRating(a, p) * 1.0 + Retrieve.ratioOfSquads(a, p) * 1.0 + ratioOfStrength * 6.0 + Retrieve.detachmentCheck(a, p) * 3.0) / 11;
        if (similarity >= 0.673) {
            CBRMain.reuseExplanations.add("The retrieved army has a high chance of winning. No adaption is necessary.");
            return a;
        }

        /*
            If the similarity was not greater than 0.673 then the adaption process starts.
            In each stage we check to see if the retrieved army has a worse ratio than the problem army.
            If it does, we accommodate that stage appropriately.
         */

        //Create an army list to hold various iterations of the army.
        ArrayList<Army> solutions = new ArrayList<>();
        List<List<List<Integer>>> totalOptions = new ArrayList<>();
        List<List<List<String>>> totalParameters = new ArrayList<>();
        List<String> traceReuse = new ArrayList<>();
        List<Integer> efficiencyList = new ArrayList<>();


        //Copy the retrieved army, in case all of the armies end up being worse than it.
        List<Squad> bestList = new ArrayList<>();
        for (int k = 0; k < a.getSquad().size(); k++) {
            bestList.add(new Squad(a.getSquad().get(k)));
        }
        Army retrievedArmy = new Army(a.getName(), a.getCost(), a.getRating(), bestList);
        solutions.add(retrievedArmy);
        totalOptions.add(new ArrayList<>(CBRMain.options.get(CBRMain.options.size() - 1)));
        totalParameters.add(new ArrayList<>(CBRMain.parameters.get(CBRMain.parameters.size() - 1)));

        List<Double> efficiency = new ArrayList<>();
        for (int i = 0; i < a.getSquad().size(); i++) {

            double rating = rating(a.getSquad().get(i));
            double movement = CBRMain.movementWeight * movement(a.getSquad().get(i), p);
            double psychic = CBRMain.psychicWeight * psychic(a.getSquad().get(i), p);
            double shooting = CBRMain.shootingWeight * shooting(a.getSquad().get(i), p);
            double assault = CBRMain.assaultWeight * assault(a.getSquad().get(i), p);
            double total = rating + movement + psychic + shooting + assault;

            efficiency.add(total);
            efficiencyList.add(0);
        }

        /*
            Adapt each squad through three steps.
            First step is check if the squads removal messes up formations. If it does, skip the squad.
            If it does not, pick a squad that from the SquadList that is better than said squad within the same cost.
            If none can be found, try to transform the squad by iterating through all the options.
            If that does not work (no points or no increase) remove squad from army and continue from the next squad.
            We should arrive at a higher total solution at the end of this process.
         */

        int i = 0;
        int efficiencyCounter = 0;
        traceReuse.add("An attempt will be made to improve the army, by looking at individual squads.");
        traceReuse.add("A total usefulness of each individual squad is their rating and efficiency in each phase.");
        traceReuse.add("The efficiency of a unit is calculated by comparing the unit to the enemy army, and seeing how it would perform.");
        traceReuse.add("As there are many variables that come into calculation, not all of them can be presented. Information on what variables are taken into account is presented.");
        traceReuse.add("");
        traceReuse.add("Rating is the performance of squads from the start of the systems usage. The higher the rating the better a squad is performing.");
        traceReuse.add("The starting (default) rating for all squads is 1500.");
        traceReuse.add("");
        traceReuse.add("Movement efficiency depends on unit type and terrain.");
        traceReuse.add("");
        traceReuse.add("Psychic efficiency depends on Psyker, Psychic Pilot or Brotherhood of Psykers abilities and the levels of those abilities.");
        traceReuse.add("");
        traceReuse.add("Shooting efficiency primarily depends on the Ballistics Skill (BS).");
        traceReuse.add("All of the weapons a unit has are weighted in in comparison to the enemy units armour and armour saves.");
        traceReuse.add("For units with armour, frontal and side armour are averaged together, as it is assumed this is where the shooting will be directed from.");
        traceReuse.add("Weapons have different strengths and armour penetration values. Weapons that on average penetrate the armour of the enemy will be weighted much higher.");
        traceReuse.add("");
        traceReuse.add("Assault efficiency depends on a lot of factors, including the Weapon Skill (WS), Strength (S), number of attacks (A) and initiative (I) of a unit.");
        traceReuse.add("Initiative is highly weighted here, as units with higher initiatives may kill an enemy unit in the assault phase before the enemy unit can react.");
        traceReuse.add("All of the weapons a unit has are weighted in in comparison to the enemy units armour and armour saves.");
        traceReuse.add("For units with armour, the (usually) weaker rear armour is counted in assault, as per rules.");
        traceReuse.add("Therefore armoured units tend to be at a disadvantage in the assault phase if they are also not mobile or can not fight back.");
        traceReuse.add("Weapons have different strengths and armour penetration values. Weapons that on average penetrate the armour of the enemy will be weighted much higher.");
        traceReuse.add("");
        traceReuse.add("In the case a suitable substitution can not be found, a transformation will be applied to the squad.");
        traceReuse.add("This involves either adding or replacing a piece of equipment, adding a unit or upgrading a unit. Consult the army lists to see the exact changes");
        traceReuse.add("");
        traceReuse.add("All detachments will be preserved during this process.");
        traceReuse.add("Any numbers are automatically rounded to be more presentable.");
        traceReuse.add("");

        /*
            A varied number of iterations is performed. The more iterations the closer we are to exhausting all the solutions.
            However, performance will take a hit the more iterations we perform.
         */

        Squad sqd = null;

        while (efficiency.size() != efficiencyCounter) {

            efficiencyCounter = 0;
            boolean deletion = false;
            double lowest = Double.MAX_VALUE;
            int low = -1;
            double cost = 0;
            double points = Retrieve.points;

            for (int j = 0; j < a.getSquad().size(); j++) {
                cost += a.getSquad().get(j).getCost();
            }

            //Find the lowest total from all of the totals
            for (int j = 0; j < efficiency.size(); j++) {
                if (lowest > efficiency.get(j) && efficiencyList.get(j) != 1) {
                    lowest = efficiency.get(j);
                    low = j;
                }
            }

            if (!CaseBaseLoading.checkDuplicate(sqd, a.getSquad().get(low))) {
                traceReuse.add("Current least efficient squad is: " + a.getSquad().get(low).getName() + ". It has a calculated efficiency rating of " + String.format("%.4g", efficiency.get(low)) + ".");
                traceReuse.add("");
            }

            sqd = a.getSquad().get(low);

            //Set the temporary squad if anything goes wrong.
            Squad t = new Squad(a.getSquad().get(low));
            int detach = 0;

            //Try to perform reuse, checking for validity across each step.
            for (int j = 0; j < CBRMain.SquadList.size(); j++) {
                if (a.getSquad().get(low).getRole().equals(CBRMain.SquadList.get(j).getRole())
                        && a.getSquad().get(low).getUnit().get(0).getFaction().equals(CBRMain.SquadList.get(j).getUnit().get(0).getFaction())) {
                    a.getSquad().set(low, CBRMain.SquadList.get(j));
                    detach = GeneralKnowledge.detachments(a.getSquad().get(low).getUnit().get(0).getFaction(), a) / a.getSquad().size();
                    a.getSquad().set(low, t);
                    break;
                }
            }

            if (detach == 0) {
                traceReuse.add("Can not remove the squad since it would violate the detachment, moving on to another squad.");
                traceReuse.add("");
            }

            //If the unit does not ruin detachments, then we can try to change it.
            if (detach == 1) {

                double bestTotal = efficiency.get(low);
                Squad bestSquad = t;

                List<List<Integer>> options = CBRMain.options.get(CBRMain.options.size() - 1);
                List<Integer> squadOptions;

                List<List<String>> parameters = CBRMain.parameters.get(CBRMain.parameters.size() - 1);
                List<String> squadParameters;

                for (int j = 0; j < CBRMain.SquadList.size(); j++) {
                    if (a.getSquad().get(low).getRole().equals(CBRMain.SquadList.get(j).getRole())
                            && a.getSquad().get(low).getUnit().get(0).getFaction().equals(CBRMain.SquadList.get(j).getUnit().get(0).getFaction())
                            && a.getSquad().get(low).getCost() + (points - cost) >= CBRMain.SquadList.get(j).getCost()) {

                        Squad x = CBRMain.SquadList.get(j);

                        double rating = rating(x);
                        double movement = CBRMain.movementWeight * movement(x, p);
                        double psychic = CBRMain.psychicWeight * psychic(x, p);
                        double shooting = CBRMain.shootingWeight * shooting(x, p);
                        double assault = CBRMain.assaultWeight * assault(x, p);
                        double total = rating + movement + psychic + shooting + assault;

                        if (total > bestTotal) {
                            bestTotal = total;
                            bestSquad = new Squad(x);
                            squadOptions = CBRMain.squadOptions.get(j);
                            squadParameters = CBRMain.squadParameters.get(j);
                            options.set(low, squadOptions);
                            parameters.set(low, squadParameters);
                        }

                    }
                }

                CBRMain.options.set(CBRMain.options.size() - 1, options);
                CBRMain.parameters.set(CBRMain.parameters.size() - 1, parameters);

                if (!bestSquad.equals(t)) {

                    //If the squad has been changed, start the efficiency list empty again.
                    for (int z = 0; z < efficiencyList.size(); z++) {
                        efficiencyList.set(z, 0);
                    }

                    traceReuse.add("Squad successfully replaced with: " + bestSquad.getName() + ", which has a efficiency rating of: " + String.format("%.4g", bestTotal) + ".");
                    traceReuse.add("");

                    double rating = rating(a.getSquad().get(low));
                    double rating2 = rating(bestSquad);
                    double movement = CBRMain.movementWeight * movement(a.getSquad().get(low), p);
                    double psychic = CBRMain.psychicWeight * psychic(a.getSquad().get(low), p);
                    double shooting = CBRMain.shootingWeight * shooting(a.getSquad().get(low), p);
                    double assault = CBRMain.assaultWeight * assault(a.getSquad().get(low), p);
                    double movement2 = CBRMain.movementWeight * movement(bestSquad, p);
                    double psychic2 = CBRMain.psychicWeight * psychic(bestSquad, p);
                    double shooting2 = CBRMain.shootingWeight * shooting(bestSquad, p);
                    double assault2 = CBRMain.assaultWeight * assault(bestSquad, p);

                    if (rating != rating2) {
                        if (rating > rating2) {
                            traceReuse.add("Squad " + a.getSquad().get(low).getName() + ", has a higher rating which stands at " + String.format("%.4g", (a.getSquad().get(low).getRating())) + " versus " + bestSquad.getName() + ", which has a rating of " + String.format("%.4g", (bestSquad.getRating())) + ".");
                        } else {
                            traceReuse.add("Squad " + bestSquad.getName() + ", has a higher which stands at " + String.format("%.4g", (bestSquad.getRating())) + " versus " + a.getSquad().get(low).getName() + ", which has a rating of " + String.format("%.4g", (a.getSquad().get(low).getRating())) + ".");
                        }
                    }
                    if (movement != movement2) {
                        if (movement > movement2) {
                            traceReuse.add("Squad " + a.getSquad().get(low).getName() + ", has a better calculated movement phase efficiency by " + String.format("%.2g", (movement - movement2)) + ".");
                        } else {
                            traceReuse.add("Squad " + bestSquad.getName() + ", has a better calculated movement phase efficiency by " + String.format("%.2g", (movement2 - movement)) + ".");
                        }
                    } else {
                        traceReuse.add("The two squads are equal in movement efficiency.");
                    }
                    if (psychic != psychic2) {
                        if (psychic > psychic2) {
                            traceReuse.add("Squad " + a.getSquad().get(low).getName() + ", has a better calculated psychic phase efficiency by " + String.format("%.2g", (psychic - psychic2)) + ".");
                        } else {
                            traceReuse.add("Squad " + bestSquad.getName() + ", has a better calculated psychic phase efficiency by " + String.format("%.2g", (psychic2 - psychic)) + ".");
                        }
                    } else {
                        traceReuse.add("The two squads are equal in psychic efficiency.");
                    }
                    if (shooting != shooting2) {
                        if (shooting > shooting2) {
                            traceReuse.add("Squad " + a.getSquad().get(low).getName() + ", has a better calculated shooting phase efficiency by " + String.format("%.2g", (shooting - shooting2)) + ".");
                        } else {
                            traceReuse.add("Squad " + bestSquad.getName() + ", has a better calculated shooting phase efficiency by " + String.format("%.2g", (shooting2 - shooting)) + ".");
                        }
                    } else {
                        traceReuse.add("The two squads are equal in shooting efficiency.");
                    }
                    if (assault != assault2) {
                        if (assault > assault2) {
                            traceReuse.add("Squad " + a.getSquad().get(low).getName() + ", has a better calculated assault phase efficiency by " + String.format("%.2g", (assault - assault2)) + ".");
                        } else {
                            traceReuse.add("Squad " + bestSquad.getName() + ", has a better calculated assault phase efficiency by " + String.format("%.2g", (assault2 - assault)) + ".");
                        }
                    } else {
                        traceReuse.add("The two squads are equal in assault efficiency.");
                    }

                    traceReuse.add("");
                }

                a.getSquad().set(low, bestSquad);
                efficiency.set(low, bestTotal);

                //The squad was not substituted. Therefore, we attempt to do transformation on the squad.
                int bestOption = 0;
                if (bestSquad.equals(t)) {
                    //Loops through each unit
                    for (int j = 0; j < a.getSquad().get(low).getUnit().size(); j++) {
                        //Loops through each option of the unit
                        for (int k = 0; k < a.getSquad().get(low).getUnit().get(j).getOptions().size(); k++) {
                            //Do logic to try to use options.
                            Squad x = new Squad(a.getSquad().get(low));
                            GeneralKnowledge.Option(x.getUnit().get(j).getOptions().get(k), "None", x);
                            double tempCost = 0;
                            for (int z = 0; z < a.getSquad().size(); z++) {
                                if (z == low) {
                                    tempCost += x.getCost();
                                } else {
                                    tempCost += a.getSquad().get(z).getCost();
                                }
                            }
                            //See if we have a legal costing army.
                            if (tempCost <= points) {

                                //Has the efficiency changed?
                                double rating = rating(x);
                                double movement = CBRMain.movementWeight * movement(x, p);
                                double psychic = CBRMain.psychicWeight * psychic(x, p);
                                double shooting = CBRMain.shootingWeight * shooting(x, p);
                                double assault = CBRMain.assaultWeight * assault(x, p);
                                double total = rating + movement + psychic + shooting + assault;

                                //If the efficiency is better, set the cost total to this.
                                if (total > bestTotal) {
                                    bestSquad = new Squad(x);
                                    bestTotal = total;
                                    bestOption = x.getUnit().get(j).getOptions().get(k);
                                }
                            }
                        }
                    }
                }

                if (!bestSquad.equals(t) && bestOption != 0) {

                    for (int z = 0; z < efficiencyList.size(); z++) {
                        efficiencyList.set(z, 0);
                    }

                    traceReuse.add("Squad successfully transformed. Option " + bestOption + " was applied to the squad: " + bestSquad.getName() + ", which increased its efficiency to: " + String.format("%.4g", bestTotal));
                    traceReuse.add("");
                }

                a.getSquad().set(low, bestSquad);

                //Keep track of which options and parameters we add, so that we may retain the army correctly.
                if (!bestSquad.equals(t) && bestOption != 0) {

                    List<List<Integer>> tmp = new ArrayList<>(CBRMain.options.get(CBRMain.options.size() - 1));
                    List<Integer> tmp2 = new ArrayList<>(tmp.get(low));
                    tmp2.add(bestOption);
                    tmp.set(low, tmp2);

                    List<List<String>> partmp = new ArrayList<>(CBRMain.parameters.get(CBRMain.parameters.size() - 1));
                    List<String> partmp2 = new ArrayList<>(partmp.get(low));
                    partmp2.add("None");
                    partmp.set(low, partmp2);

                    CBRMain.options.set(CBRMain.options.size() - 1, tmp);
                    CBRMain.parameters.set(CBRMain.parameters.size() - 1, partmp);
                }

                efficiency.set(low, bestTotal);

                /*
                    Substitution and transformation have failed, either because we can not substitute and transform, or because we do not have the points to do so.
                    Therefore we need to delete a unit safely to free up some points.
                    Points should only be freed up if the point cost is close to the limit or on the limit.
                 */

                if (bestSquad.equals(t)) {
                    //First we check if the squad can be safely removed from the army
                    String faction = a.getSquad().get(low).getUnit().get(0).getFaction();
                    a.getSquad().remove(low);
                    detach = GeneralKnowledge.detachments(faction, a) / a.getSquad().size();
                    a.getSquad().add(low, t);

                    //We can remove this squad if it maintains detachment after removal.
                    if (detach == 1) {
                        //We perform removal of squad and efficiency. We only remove if our points are close to the limit. The limit is 1% of the points, rounded up. This is 20 for most armies.
                        double tempCost = 20;
                        if (points - tempCost <= cost) {
                            //We save the army before deletion, assuming this maximum is the highest maximum we can reach.
                            List<Squad> solutionList = new ArrayList<>();
                            for (int k = 0; k < a.getSquad().size(); k++) {
                                solutionList.add(new Squad(a.getSquad().get(k)));
                            }
                            traceReuse.add("Attempting to delete the squad to try to find another answer by freeing up some points.");
                            traceReuse.add("Saving the previous army composition, in the event that future adaptions prove less efficient as a whole.");
                            traceReuse.add("");
                            i++;
                            solutions.add(new Army(a.getName() + " Iteration Number: " + i, cost, a.getRating(), solutionList));
                            totalOptions.add(new ArrayList<>(CBRMain.options.get(CBRMain.options.size() - 1)));
                            totalParameters.add(new ArrayList<>(CBRMain.parameters.get(CBRMain.parameters.size() - 1)));
                            CBRMain.options.get(CBRMain.options.size() - 1).remove(low);
                            CBRMain.parameters.get(CBRMain.parameters.size() - 1).remove(low);
                            a.getSquad().remove(low);
                            efficiency.remove(low);
                            efficiencyList.remove(low);
                            deletion = true;
                        }
                    }
                }
            }

            //To prevent looping over the same lowest squad, if the lowest squad would ruin the detachment, simply remove it from the list.
            if (!deletion) {
                if (CaseBaseLoading.checkDuplicate(sqd, a.getSquad().get(low))) {
                    //Here we fill up the list of efficiencies.
                    efficiencyList.set(low, 1);
                    traceReuse.add("Skipping Squad: " + a.getSquad().get(low).getName() + ", as no improvements can be made at this moment.");
                    traceReuse.add("");
                }
            }

            //Finally, we must calculate if the loop is concluded, by seeing if the efficiency list is completely filled with 1s.
            for (Integer anEfficiencyList : efficiencyList) {
                if (anEfficiencyList == 1) {
                    efficiencyCounter++;
                }
            }
        }

        traceReuse.add("No further improvements could be found.");

        //Add the last iteration to the list of solutions.
        solutions.add(a);
        //Add the retrieved army to the list of solutions, in case the retrieval performance actually lowerd the similarity.
        totalOptions.add(new ArrayList<>(CBRMain.options.get(CBRMain.options.size() - 1)));
        totalParameters.add(new ArrayList<>(CBRMain.parameters.get(CBRMain.parameters.size() - 1)));

        Army returnArmy = Retrieve.RetrieveBest(solutions, p);

        if (CBRMain.winChanceAfterRevise == 0) {
            if (CBRMain.winChanceBeforeReuse != CBRMain.winChanceAfterReuse) {
                CBRMain.reuseExplanations.add("The army has been adapted. The new army has a " + String.format("%.4g", CBRMain.winChanceAfterReuse * 100.0) + "% chance to win this match.");
                for (int j = 0; j < traceReuse.size(); j++) {
                    CBRMain.reuseExplanations.add(traceReuse.get(j));
                }
            } else {
                CBRMain.reuseExplanations.add("After exhausting all the options, the original army remains the best choice. No adaption was performed.");
                for (int j = 0; j < traceReuse.size(); j++) {
                    CBRMain.reuseExplanations.add(traceReuse.get(j));
                }
            }
        } else {
            if (CBRMain.winChanceAfterReuse != CBRMain.winChanceAfterRevise) {
                CBRMain.reviseExplanations.add("The army has been revised with the new ratings in mind. The new army has a " + String.format("%.4g", CBRMain.winChanceAfterRevise * 100.0) + "% chance to win this match.");
                for (int j = 0; j < traceReuse.size(); j++) {
                    CBRMain.reviseExplanations.add(traceReuse.get(j));
                }
            } else {
                CBRMain.reviseExplanations.add("No revision was possible, as there were no sufficient changes to warrant it.");
                for (int j = 0; j < traceReuse.size(); j++) {
                    CBRMain.reviseExplanations.add(traceReuse.get(j));
                }
            }
        }

        for (int j = 0; j < solutions.size(); j++) {
            if (solutions.get(j).equals(returnArmy)) {
                CBRMain.options.set(CBRMain.options.size() - 1, totalOptions.get(j));
                CBRMain.parameters.set(CBRMain.parameters.size() - 1, totalParameters.get(j));
            }
        }


        for (int j = 0; j < a.getSquad().size(); j++) {

            //Add the armies squads to the Squad Case Base
            int countOccurences = 1;
            for (int k = 0; k < CBRMain.SquadList.size(); k++) {
                if (CBRMain.SquadList.get(k).getName().contains(a.getSquad().get(j).getName())) {
                    countOccurences++;
                }
            }
            int countDuplicates = 0;

            //Make insert a new squad (so that the naming is preserved for the army).
            Squad insert = new Squad(a.getSquad().get(j));

            //Set the name and the base of the insert to be different, so that it is a different case.
            insert.setName(a.getSquad().get(j).getName() + " " + countOccurences);
            insert.setRating(1500.0);
            insert.setBase(false);

            for (int k = 0; k < CBRMain.SquadList.size(); k++) {
                if (CBRMain.SquadList.get(k).getName().contains(a.getSquad().get(j).getName())) {
                    //Check if there is a single duplicate in the system. If there is, we do not add a squad.
                    if (CaseBaseLoading.checkDuplicate(CBRMain.SquadList.get(k), insert)) {
                        countDuplicates++;
                    }
                }
            }

            if (countDuplicates == 0) {
                CBRMain.SquadList.add(insert);
                CBRMain.squadOptions.add(CBRMain.options.get(CBRMain.options.size() - 1).get(j));
                CBRMain.squadParameters.add(CBRMain.parameters.get(CBRMain.parameters.size() - 1).get(j));
            }
        }

        return returnArmy;
    }

    //Compares the movement of the squad to the average of the problem army.
    public static double movement(Squad s, Army p){

        int unitCount = 0;
        int transportCount = 0;
        int infantryCount = 0;
        double normalMovement = 0;
        double expeditousMovement = 0;
        double similarity = 0;

        double finalNormal;
        double finalExpeditous;

        /*
           Uses general domain knowledge to check for movement speeds, but also checks for transports and infantry.
           It is assumed that transports are always used to the best of ones ability (that is to say, to transport infantry).
           The type of transport is (currently) ignored.
         */
        for (int i = 0; i < p.getSquad().size(); i++){
            for (int j = 0; j < p.getSquad().get(i).getUnit().size(); j++){
                if (p.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    normalMovement += GeneralKnowledge.movementKnowledge(p.getSquad().get(i).getUnit().get(j).getType(), true);
                    expeditousMovement += GeneralKnowledge.movementKnowledge(p.getSquad().get(i).getUnit().get(j).getType(), false);
                    unitCount++;
                    if (p.getSquad().get(i).getUnit().get(j).getType().contains("Transport")) {
                        transportCount++;
                    }
                    if (p.getSquad().get(i).getUnit().get(0).getType().contains("Infantry")) {
                        infantryCount++;
                    }
                }
            }
        }

        expeditousMovement += normalMovement;

        if ((infantryCount > transportCount) && (transportCount != 0)){
            normalMovement = normalMovement + infantryCount*3;
        }
        else{
            normalMovement = normalMovement + transportCount*3;
        }

        normalMovement = normalMovement/unitCount;
        expeditousMovement = expeditousMovement/unitCount;

        finalNormal = normalMovement*Retrieve.terrainRate(p, terrain);
        finalExpeditous = expeditousMovement*Retrieve.terrainRate(p, terrain);

        //Do the same for solution army
        normalMovement = 0;
        expeditousMovement = 0;
        unitCount = 0;
        infantryCount = 0;
        transportCount = 0;

        for (int j = 0; j < s.getUnit().size(); j++) {
            if (s.getUnit().get(j).getCount() > 0) {
                normalMovement += GeneralKnowledge.movementKnowledge(s.getUnit().get(j).getType(), true);
                expeditousMovement += GeneralKnowledge.movementKnowledge(s.getUnit().get(j).getType(), false);
                unitCount++;
                if (s.getUnit().get(j).getType().contains("Transport")) {
                    transportCount++;
                }
                if (s.getUnit().get(0).getType().contains("Infantry")) {
                    infantryCount++;
                }
            }
        }


        expeditousMovement += normalMovement;

        if ((infantryCount > transportCount) && (transportCount != 0)){
            normalMovement = normalMovement + infantryCount*3;
        }
        else{
            normalMovement = normalMovement + transportCount*3;
        }

        normalMovement = normalMovement/unitCount;
        expeditousMovement = expeditousMovement/unitCount;

        finalNormal -= normalMovement*Retrieve.terrainRate(p, terrain);
        finalExpeditous -= expeditousMovement*Retrieve.terrainRate(p, terrain);

        /*
            Currently, normal movement is weighted at 70% while expeditious is weighted at 30%.
            If an entire army on average can move 1 inch move, it is much more maneuverable than the opposing army.
         */
        if (finalNormal < -3){
            similarity += 0.9;
        }

        if (finalExpeditous < -3){
            similarity += 0.1;
        }

        if (finalNormal < 3 && finalNormal > -3){
            similarity += 0.9*(0.5+0.166*(-finalNormal));
        }

        if (finalExpeditous < 3 && finalExpeditous > -3){
            similarity += 0.1*(0.5+0.166*(-finalExpeditous));
        }

        return similarity;
    }

    //Compares the psychic strengths of squads to that of the enemy army average
    public static double psychic(Squad s, Army p){

        double similarity = 0;
        int problemPsychic = 0;
        int solutionPsychic = 0;

        /*
           Counts the amount of Psychic mastery levels each army has and compares them.
           The psychic power ratio is just a measure of psychic strength, it is up to the players to use it correctly.
         */

        for (int i = 0; i < p.getSquad().size(); i++){
            for (int j = 0; j < p.getSquad().get(i).getUnit().size(); j++){
                if (p.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    for (int k = 0; k < p.getSquad().get(i).getUnit().get(j).getRules().size(); k++) {
                        if (p.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Psyker") ||
                                p.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Psychic Pilot") ||
                                p.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Brotherhood of Psykers") ||
                                p.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Brotherhood of Sorcerers")) {
                            problemPsychic += Integer.parseInt(p.getSquad().get(i).getUnit().get(j).getRules().get(k).substring(p.getSquad().get(i).getUnit().get(j).getRules().get(k).length() - 1, p.getSquad().get(i).getUnit().get(j).getRules().get(k).length()));
                        }
                    }
                }
            }
        }

        for (int j = 0; j < s.getUnit().size(); j++) {
            if (s.getUnit().get(j).getCount() > 0) {
                for (int k = 0; k < s.getUnit().get(j).getRules().size(); k++) {
                    if (s.getUnit().get(j).getRules().get(k).contains("Psyker") ||
                            s.getUnit().get(j).getRules().get(k).contains("Psychic Pilot") ||
                            s.getUnit().get(j).getRules().get(k).contains("Brotherhood of Psykers") ||
                           s.getUnit().get(j).getRules().get(k).contains("Brotherhood of Sorcerers")) {
                        solutionPsychic += Integer.parseInt(s.getUnit().get(j).getRules().get(k).substring(s.getUnit().get(j).getRules().get(k).length() - 1, s.getUnit().get(j).getRules().get(k).length()));
                    }
                }
            }
        }


        //in the case of no psychic abilities the armies are tied, and similarity is 0.5.
        if (solutionPsychic == 0 && problemPsychic == 0){
            similarity = 0.5;
        }
        //If the difference is 3 or more, an army has a major advantage
        else if (solutionPsychic - problemPsychic >= 3){
            similarity = 1.0;
        }
        else if (solutionPsychic - problemPsychic > -3 && solutionPsychic - problemPsychic < 3){
            similarity = ((solutionPsychic - problemPsychic)*(0.16666))+0.5;
        }

        return similarity;
    }

    public static double shooting(Squad s, Army p){

        double proavgToughness = 0;
        double proavgAS = 0;
        double solBarrage = 0;
        int vehC = 0;
        int infC = 0;

        double ballistics;
        double hitChance;
        double solutionEfficiency = 0;

        //First we determine the average toughness, armour and armour saves for the problem and solution armies

        for (int i = 0; i < p.getSquad().size(); i++) {
            for (int j = 0; j < p.getSquad().get(i).getUnit().size(); j++) {
                if (p.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    if (p.getSquad().get(i).getUnit().get(j).getType().equals("Infantry") || p.getSquad().get(i).getUnit().get(j).getType().equals("Infantry-Character")
                            || p.getSquad().get(i).getUnit().get(j).getType().equals("Bike") || p.getSquad().get(i).getUnit().get(j).getType().equals("Bike-Character")
                            || p.getSquad().get(i).getUnit().get(j).getType().equals("Artillery")) {
                        infC+= p.getSquad().get(i).getUnit().get(j).getCount();
                        proavgAS += p.getSquad().get(i).getUnit().get(j).getAS()*p.getSquad().get(i).getUnit().get(j).getCount();
                        proavgToughness += p.getSquad().get(i).getUnit().get(j).getT()*p.getSquad().get(i).getUnit().get(j).getCount();
                    } else {
                        vehC+= p.getSquad().get(i).getUnit().get(j).getCount();
                        proavgToughness += ((p.getSquad().get(i).getUnit().get(j).getSide()+p.getSquad().get(i).getUnit().get(j).getSide())/2)*p.getSquad().get(i).getUnit().get(j).getCount();
                    }
                }
            }
        }

        if (infC > 0) proavgAS = proavgAS / infC;
        if (infC > 0 || vehC > 0) proavgToughness = proavgToughness / (infC+vehC);

        //We determine the approximate effectiveness for solution squads.
        for (int j = 0; j < s.getUnit().size(); j++) {
            if (s.getUnit().get(j).getCount() > 0) {

                ballistics = s.getUnit().get(j).getBS();
                if (ballistics == 0 && s.getUnit().get(j).getType().equals("Artillery")) {
                    ballistics = s.getUnit().get(j + 1).getBS();
                }
                List<Double> effectiveness = new ArrayList<>();
                List<String> weaponNames = new ArrayList<>();
                int unitC = (int) s.getUnit().get(j).getCount();

                if (ballistics < 6) {
                    hitChance = ballistics / 6.0;
                } else {
                    hitChance = (5.0 / 6.0) + ((ballistics - 5.0) / 36.0);
                }

                for (int k = 0; k < s.getUnit().get(j).getEquipment().size(); k++) {
                    for (int x = 0; x < CBRMain.EquipmentList.size(); x++) {
                        if (s.getUnit().get(j).getEquipment().get(k).getName().equals(CBRMain.EquipmentList.get(x).getName())) {
                            if (CBRMain.EquipmentList.get(x).getRange() > 0 || CBRMain.EquipmentList.get(x).getRange() == -1) {

                                double r = CBRMain.EquipmentList.get(x).getRange();
                                int str = Integer.parseInt(CBRMain.EquipmentList.get(x).getStrength());
                                double armourp = CBRMain.EquipmentList.get(x).getAp();
                                double effect;
                                double armourEffect;
                                double rangeEffect;
                                double strengtheffect;

                                if (r < 12) rangeEffect = 0.25;
                                else if (r >= 12 && r < 24) rangeEffect = 0.5;
                                else if (r >= 24 && r < 36) rangeEffect = 1.0;
                                else rangeEffect = 2.0;

                                if (CBRMain.EquipmentList.get(x).getType().contains("Twin-linked")) {
                                    hitChance = hitChance + ((1 - hitChance) * (hitChance));
                                }

                                if (r == -1) {
                                    hitChance = 1.0;
                                }

                                //This weapon can not wound (and furthermore can not even dent armour), so the effectiveness is 0.
                                if (str - proavgToughness < -2) {
                                    strengtheffect = 0;
                                }
                                //This type of weapon is much stronger than toughness.
                                else if (str - proavgToughness >= 2) {
                                    strengtheffect = 5.0 / 6.0;
                                }
                                //This type of weapon is in between the above two values.
                                else {
                                    strengtheffect = (str - proavgToughness) * 0.1665 + 0.5;
                                }

                                if (armourp <= proavgAS && armourp != 0) {
                                    armourEffect = 1.0;
                                } else {
                                    armourEffect = 1 - ((7.0 - proavgAS) / 6.0);
                                }

                                effect = hitChance * strengtheffect * armourEffect * rangeEffect;
                                effectiveness.add(effect * GeneralKnowledge.hitCount(CBRMain.EquipmentList.get(x).getType()));
                                weaponNames.add(CBRMain.EquipmentList.get(x).getName());
                            }
                        }
                    }
                }
                if (!effectiveness.isEmpty()) {
                    int sort = 0;
                    while (sort < unitC) {
                        double best = -0.01;
                        int tmp = -1;
                        for (int k = 0; k < effectiveness.size(); k++) {
                            if (effectiveness.get(k) > best) {
                                best = effectiveness.get(k);
                                tmp = k;
                            }
                        }
                        for (int k = 0; k < s.getUnit().get(j).getEquipment().size(); k++) {
                            if (weaponNames.get(tmp).equals(s.getUnit().get(j).getEquipment().get(k).getName())) {
                                solutionEfficiency += effectiveness.get(tmp) * s.getUnit().get(j).getEquipment().get(k).getCount();
                                sort += s.getUnit().get(j).getEquipment().get(k).getCount();
                                for (int x = 0; x < CBRMain.EquipmentList.size(); x++) {
                                    if (weaponNames.get(tmp).equals(CBRMain.EquipmentList.get(x).getName()) && CBRMain.EquipmentList.get(x).getType().contains("Barrage")) {
                                        solBarrage += effectiveness.get(tmp) * s.getUnit().get(j).getEquipment().get(k).getCount();
                                    }
                                }
                                effectiveness.remove(tmp);
                            }
                        }
                        if (effectiveness.isEmpty()) {
                            sort += unitC;
                        }
                    }
                }
            }
        }

        //We calculate how much cover influences our shooting (The more cover, the less our overall effectiveness)
        solutionEfficiency = solutionEfficiency - solBarrage;
        solutionEfficiency*=Retrieve.shootingWeight(terrain);
        solutionEfficiency+=solBarrage;

        //Finally we return the value for the solution efficiency.
        return solutionEfficiency;
    }

    public static double assault(Squad s, Army p){

        double proavgToughness = 0;
        double proavgAS = 0;
        double proavgWS = 0;
        int vehC = 0;
        int infC = 0;

        double weaponskill;
        double strength;
        double attack;
        double hitChance = 0;
        double solutionEfficiency = 0;

        //First we determine the average toughness, armour and armour saves for the problem and solution armies

        for (int i = 0; i < p.getSquad().size(); i++) {
            for (int j = 0; j < p.getSquad().get(i).getUnit().size(); j++) {
                if (p.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    if (p.getSquad().get(i).getUnit().get(j).getType().equals("Infantry") || p.getSquad().get(i).getUnit().get(j).getType().equals("Infantry-Character")
                            || p.getSquad().get(i).getUnit().get(j).getType().equals("Bike") || p.getSquad().get(i).getUnit().get(j).getType().equals("Bike-Character")
                            || p.getSquad().get(i).getUnit().get(j).getType().equals("Artillery")) {
                        infC+= p.getSquad().get(i).getUnit().get(j).getCount();
                        proavgAS += p.getSquad().get(i).getUnit().get(j).getAS()*p.getSquad().get(i).getUnit().get(j).getCount();
                        proavgToughness += p.getSquad().get(i).getUnit().get(j).getT()*p.getSquad().get(i).getUnit().get(j).getCount();
                        proavgWS += p.getSquad().get(i).getUnit().get(j).getWS()*p.getSquad().get(i).getUnit().get(j).getCount();
                    } else {
                        vehC+= p.getSquad().get(i).getUnit().get(j).getCount();
                        if (p.getSquad().get(i).getUnit().get(j).getWS() == 0){
                            proavgWS+=p.getSquad().get(i).getUnit().get(j).getCount();
                        }
                        else proavgWS += p.getSquad().get(i).getUnit().get(j).getWS()*p.getSquad().get(i).getUnit().get(j).getCount();
                        proavgToughness += p.getSquad().get(i).getUnit().get(j).getRear()*p.getSquad().get(i).getUnit().get(j).getCount();
                    }
                }
            }
        }

        if (infC > 0) proavgAS = proavgAS / infC;
        if (infC > 0 || vehC > 0) proavgToughness = proavgToughness / (infC+vehC);
        proavgWS = proavgWS / (infC+vehC);

        //We determine the approximate effectiveness for solution squad.
        for (int j = 0; j < s.getUnit().size(); j++) {
            if (s.getUnit().get(j).getCount() > 0) {

                weaponskill = s.getUnit().get(j).getWS();
                strength = s.getUnit().get(j).getS();
                List<Double> effectiveness = new ArrayList<>();
                List<String> weaponNames = new ArrayList<>();
                int unitC = (int) s.getUnit().get(j).getCount();

                //calculate to hit
                if (weaponskill > proavgWS) {
                    hitChance = 4.0 / 6.0;
                } else if ((weaponskill == proavgWS) || (weaponskill < proavgWS && weaponskill >= 5) || (weaponskill == 1 && proavgWS < 2)
                        || (weaponskill == 2 && proavgWS < 5) || (weaponskill == 3 && proavgWS < 7)
                        || (weaponskill == 4 && proavgWS < 9)) {
                    hitChance = 3.0 / 6.0;
                } else if ((weaponskill == 1 && proavgWS > 2) || (weaponskill == 2 && proavgWS > 5)
                        || (weaponskill == 3 && proavgWS > 7) || (weaponskill == 4 && proavgWS > 9)) {
                    hitChance = 2.0 / 6.0;
                }

                //Calculate effectiveness for weapons
                for (int k = 0; k < s.getUnit().get(j).getEquipment().size(); k++) {
                    for (int x = 0; x < CBRMain.EquipmentList.size(); x++) {
                        if (s.getUnit().get(j).getEquipment().get(k).getName().equals(CBRMain.EquipmentList.get(x).getName())) {
                            if (CBRMain.EquipmentList.get(x).getRange() == 0) {

                                double effect;
                                String str = CBRMain.EquipmentList.get(x).getStrength();
                                double armourp = CBRMain.EquipmentList.get(x).getAp();
                                double armourEffect;
                                double strengthEffect = 0;
                                double initiativeeffect = 0.1 * s.getUnit().get(j).getI();


                                if (initiativeeffect == 0) {
                                    initiativeeffect = 0.1;
                                }

                                if (str.equals("+1")) {
                                    strength++;
                                }
                                if (str.equals("+2")) {
                                    strength += 2;
                                }
                                if (str.equals("+6")) {
                                    strength += 6;
                                }
                                if (str.equals("x2")) {
                                    strength *= 2;
                                }

                                if (strength > 10) {
                                    strength = 10;
                                }

                                //This strength difference will not cause wounds.
                                if (strength < 7 && proavgToughness > strength + 4) {
                                    strengthEffect = 0;
                                } else if (proavgToughness - strength > 2) {
                                    strengthEffect = 1.0 / 6.0;
                                } else if (proavgToughness - strength < 2 && proavgToughness - strength > 0) {
                                    strengthEffect = 2.0 / 6.0;
                                } else if (proavgToughness == strength) {
                                    strengthEffect = 3.0 / 6.0;
                                } else if (strength - proavgToughness < 2 && strength - proavgToughness > 0) {
                                    strengthEffect = 4.0 / 6.0;
                                } else if (strength - proavgToughness >= 2) {
                                    strengthEffect = 5.0 / 6.0;
                                }

                                if (armourp <= proavgAS && armourp != 0) {
                                    armourEffect = 1.0;
                                } else {
                                    armourEffect = 1 - ((7.0 - proavgAS) / 6.0);
                                }

                                for (int z = 0; z < CBRMain.EquipmentList.get(x).getType().size(); z++) {
                                    if (CBRMain.EquipmentList.get(x).getType().get(z).equals("Unwieldy")) {
                                        initiativeeffect = 0.1;
                                    }
                                }

                                effect = hitChance * strengthEffect * armourEffect * initiativeeffect;
                                effectiveness.add(effect * s.getUnit().get(j).getA());
                                weaponNames.add(CBRMain.EquipmentList.get(x).getName());

                            }
                        }
                    }
                }

                    /*
                        Check to see if the list of weapons is empty. If the list is empty, the unit has no melee weapons
                        In this case the unit is still armed with a default melee weapon (not mentioned or included in the unit).
                     */

                if (effectiveness.isEmpty()) {

                    double effect;
                    double armourp = 0;
                    double armourEffect;
                    double strengthEffect = 0;
                    double initiativeeffect = 0.1 * s.getUnit().get(j).getI();
                    attack = s.getUnit().get(j).getA();

                    if (attack == 0) {
                        attack = 1;
                    }

                    if (initiativeeffect == 0) {
                        initiativeeffect = 0.1;
                    }

                    if (strength < 7 && proavgToughness > strength + 4) {
                        strengthEffect = 0;
                    } else if (proavgToughness - strength > 2) {
                        strengthEffect = 1.0 / 6.0;
                    } else if (proavgToughness - strength < 2 && proavgToughness - strength > 0) {
                        strengthEffect = 2.0 / 6.0;
                    } else if (proavgToughness == strength) {
                        strengthEffect = 3.0 / 6.0;
                    } else if (strength - proavgToughness < 2 && strength - proavgToughness > 0) {
                        strengthEffect = 4.0 / 6.0;
                    } else if (strength - proavgToughness >= 2) {
                        strengthEffect = 5.0 / 6.0;
                    }

                    if (armourp <= proavgAS && armourp != 0) {
                        armourEffect = 1.0;
                    } else {
                        armourEffect = 1 - ((7.0 - proavgAS) / 6.0);
                    }

                    effect = hitChance * strengthEffect * armourEffect * initiativeeffect;
                    solutionEfficiency += effect * s.getUnit().get(j).getCount() * attack;
                }

                if (!effectiveness.isEmpty() && !weaponNames.isEmpty()) {
                    int sort = 0;
                    while (sort < unitC) {
                        double best = -0.01;
                        int tmp = -1;
                        for (int k = 0; k < effectiveness.size(); k++) {
                            if (effectiveness.get(k) > best) {
                                best = effectiveness.get(k);
                                tmp = k;
                            }
                        }
                        for (int k = 0; k < s.getUnit().get(j).getEquipment().size(); k++) {
                            if (weaponNames.get(tmp).equals(s.getUnit().get(j).getEquipment().get(k).getName())) {
                                solutionEfficiency += effectiveness.get(tmp) * s.getUnit().get(j).getEquipment().get(k).getCount();
                                sort += s.getUnit().get(j).getEquipment().get(k).getCount();
                                effectiveness.remove(tmp);
                            }
                        }
                        if (effectiveness.isEmpty()) {
                            sort += unitC;
                        }
                    }
                }
            }
        }

        //Finally we return the value for the problem efficiency vs. solution efficiency.
        solutionEfficiency*=Retrieve.assaultWeight(terrain);

        return solutionEfficiency;
    }

    public static double rating(Squad s){
        return Math.pow(10, (s.getRating() / 400.0)) / (Math.pow(10, (s.getRating() / 400.0)) + Math.pow(10, (1500.0 / 400.0)));
    }
}
